"""Tests for annotation."""

import unittest

import matplotlib
import numpy as np

import aspecd.annotation as annotation
import aspecd.dataset
import aspecd.exceptions
import aspecd.history
import aspecd.plotting


class TestDatasetAnnotation(unittest.TestCase):
    def setUp(self):
        self.annotation = annotation.DatasetAnnotation()
        self.annotation.content["foo"] = "bar"

    def test_instantiate_class(self):
        pass

    def test_has_scope_property(self):
        self.assertTrue(hasattr(self.annotation, "scope"))

    def test_scope_property_is_string(self):
        self.assertTrue(isinstance(self.annotation.scope, str))

    def test_has_content_property(self):
        self.assertTrue(hasattr(self.annotation, "content"))

    def test_content_property_is_dict(self):
        self.assertTrue(isinstance(self.annotation.content, dict))

    def test_has_dataset_property(self):
        self.assertTrue(hasattr(self.annotation, "dataset"))

    def test_dataset_property_is_initially_none(self):
        self.assertEqual(self.annotation.dataset, None)

    def test_has_type_property(self):
        self.assertTrue(hasattr(self.annotation, "type"))

    def test_type_property_equals_lower_classname(self):
        self.assertEqual(
            self.annotation.type, self.annotation.__class__.__name__.lower()
        )

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.annotation, "to_dict"))
        self.assertTrue(callable(self.annotation.to_dict))

    def test_to_dict_does_not_contain_certain_keys(self):
        for key in ["dataset", "type"]:
            with self.subTest(key=key):
                self.assertNotIn(key, self.annotation.to_dict())

    def test_has_annotate_method(self):
        self.assertTrue(hasattr(self.annotation, "annotate"))
        self.assertTrue(callable(self.annotation.annotate))

    def test_annotate_without_argument_and_with_dataset(self):
        self.annotation.dataset = aspecd.dataset.Dataset()
        self.annotation.annotate()
        self.assertGreater(len(self.annotation.dataset.annotations), 0)

    def test_annotate_via_dataset_annotate(self):
        test_dataset = aspecd.dataset.Dataset()
        test_dataset.annotate(self.annotation)
        self.assertGreater(len(test_dataset.annotations), 0)

    def test_annotate_with_dataset(self):
        test_dataset = aspecd.dataset.Dataset()
        self.annotation.annotate(test_dataset)
        self.assertGreater(len(test_dataset.annotations), 0)

    def test_annotate_without_argument_nor_dataset_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingDatasetError):
            self.annotation.annotate()

    def test_annotate_with_empty_content_raises(self):
        self.annotation.dataset = aspecd.dataset.Dataset()
        self.annotation.content.clear()
        with self.assertRaises(aspecd.exceptions.NoContentError):
            self.annotation.annotate()

    def test_annotate_with_empty_scope_sets_default_scope(self):
        self.annotation.dataset = aspecd.dataset.Dataset()
        self.annotation.annotate()
        self.assertEqual(self.annotation.scope, "dataset")

    def test_setting_unknown_scope_raises(self):
        with self.assertRaises(aspecd.exceptions.UnknownScopeError):
            self.annotation.scope = "foo"

    def test_annotate_returns_dataset(self):
        test_dataset = self.annotation.annotate(aspecd.dataset.Dataset())
        self.assertTrue(isinstance(test_dataset, aspecd.dataset.Dataset))

    def test_has_create_history_record_method(self):
        self.assertTrue(hasattr(self.annotation, "create_history_record"))
        self.assertTrue(callable(self.annotation.create_history_record))

    def test_create_history_record_returns_history_record(self):
        self.annotation.dataset = aspecd.dataset.Dataset()
        history_record = self.annotation.create_history_record()
        self.assertIsInstance(
            history_record, aspecd.history.AnnotationHistoryRecord
        )


class TestComment(unittest.TestCase):
    def setUp(self):
        self.comment = annotation.Comment()

    def test_instantiate_class(self):
        pass

    def test_type_property_equals_lower_classname_in_derived_class(self):
        self.assertEqual(
            self.comment.type, self.comment.__class__.__name__.lower()
        )

    def test_content_has_key_comment(self):
        self.assertTrue("comment" in self.comment.content)

    def test_has_comment_property(self):
        self.assertTrue(hasattr(self.comment, "comment"))

    def test_comment_property_sets_content_comment(self):
        commenttext = "Lorem ipsum"
        self.comment.comment = commenttext
        self.assertEqual(self.comment.content["comment"], commenttext)

    def test_comment_property_gets_content_comment(self):
        commenttext = "Lorem ipsum"
        self.comment.content["comment"] = commenttext
        self.assertEqual(self.comment.comment, commenttext)


class TestArtefact(unittest.TestCase):
    def setUp(self):
        self.artefact = annotation.Artefact()

    def test_instantiate_class(self):
        pass

    def test_content_has_key_comment(self):
        self.assertTrue("comment" in self.artefact.content)


class TestCharacteristic(unittest.TestCase):
    def setUp(self):
        self.characteristic = annotation.Characteristic()

    def test_instantiate_class(self):
        pass


class TestPlotAnnotation(unittest.TestCase):
    def setUp(self):
        self.annotation = annotation.PlotAnnotation()

    def test_instantiate_class(self):
        pass

    def test_has_plotter_property(self):
        self.assertTrue(hasattr(self.annotation, "plotter"))

    def test_plotter_property_is_initially_none(self):
        self.assertEqual(self.annotation.plotter, None)

    def test_has_type_property(self):
        self.assertTrue(hasattr(self.annotation, "type"))

    def test_type_property_equals_lower_classname(self):
        self.assertEqual(
            self.annotation.type, self.annotation.__class__.__name__.lower()
        )

    def test_has_parameters_property(self):
        self.assertTrue(hasattr(self.annotation, "parameters"))
        self.assertIsInstance(self.annotation.parameters, dict)

    def test_has_properties_property(self):
        self.assertTrue(hasattr(self.annotation, "properties"))

    def test_has_drawings_property(self):
        self.assertTrue(hasattr(self.annotation, "drawings"))
        self.assertIsInstance(self.annotation.drawings, list)

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.annotation, "to_dict"))
        self.assertTrue(callable(self.annotation.to_dict))

    def test_to_dict_does_not_contain_certain_keys(self):
        for key in ["plotter", "type", "drawings"]:
            with self.subTest(key=key):
                self.assertNotIn(key, self.annotation.to_dict())

    def test_has_annotate_method(self):
        self.assertTrue(hasattr(self.annotation, "annotate"))
        self.assertTrue(callable(self.annotation.annotate))

    def test_annotate_without_argument_and_with_plotter(self):
        self.annotation.plotter = aspecd.plotting.Plotter()
        self.annotation.annotate()
        self.assertGreater(len(self.annotation.plotter.annotations), 0)

    def test_annotate_via_plotter_annotate(self):
        test_plotter = aspecd.plotting.Plotter()
        test_plotter.annotate(self.annotation)
        self.assertGreater(len(test_plotter.annotations), 0)

    def test_annotate_with_plotter(self):
        test_plotter = aspecd.plotting.Plotter()
        self.annotation.annotate(test_plotter)
        self.assertGreater(len(test_plotter.annotations), 0)

    def test_annotate_without_argument_nor_plotter_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingPlotterError):
            self.annotation.annotate()

    @unittest.skip
    def test_annotate_with_empty_parameters_raises(self):
        self.annotation.plotter = aspecd.plotting.Plotter()
        self.annotation.parameters.clear()
        with self.assertRaises(aspecd.exceptions.NoContentError):
            self.annotation.annotate()

    def test_annotate_returns_plotter(self):
        test_plotter = self.annotation.annotate(aspecd.plotting.Plotter())
        self.assertTrue(isinstance(test_plotter, aspecd.plotting.Plotter))


class TestVerticalLine(unittest.TestCase):
    def setUp(self):
        self.annotation = annotation.VerticalLine()
        self.plotter = aspecd.plotting.Plotter()

    def test_instantiate_class(self):
        pass

    def test_annotate_adds_line_to_plotter(self):
        self.annotation.parameters["positions"] = [0.5]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_adds_line_at_correct_position(self):
        self.annotation.parameters["positions"] = [0.5]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            annotation_.parameters["positions"][0],
            annotation_.drawings[0].get_xdata()[0],
        )

    def test_annotate_adds_lines_to_plotter(self):
        self.annotation.parameters["positions"] = [0.25, 0.5, 0.75]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            len(annotation_.parameters["positions"]),
            len(annotation_.drawings),
        )
        for drawing in annotation_.drawings:
            self.assertIn(drawing, self.plotter.ax.get_children())

    def test_annotate_adds_line_to_plotter_after_plotting(self):
        self.annotation.parameters["positions"] = [0.5]
        annotation_ = self.plotter.annotate(self.annotation)
        self.plotter.plot()
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_before_plotting_does_not_add_annotation_twice(self):
        self.annotation.parameters["positions"] = [0.5]
        self.plotter.annotate(self.annotation)
        self.plotter.plot()
        self.assertEqual(1, len(self.plotter.annotations))

    def test_set_line_colour_from_dict(self):
        line_colour = "#cccccc"
        properties = {"color": line_colour}
        self.annotation.properties.from_dict(properties)
        self.assertEqual(line_colour, self.annotation.properties.color)

    def test_annotate_sets_correct_line_color(self):
        color = "#cccccc"
        properties = {"color": color}
        self.annotation.properties.from_dict(properties)
        self.plotter.plot()
        self.annotation.parameters["positions"] = [0.5]
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(color, annotation_.drawings[0].get_color())

    def test_annotate_sets_correct_line_color_for_each_line(self):
        color = "#cccccc"
        properties = {"color": color}
        self.annotation.properties.from_dict(properties)
        self.plotter.plot()
        self.annotation.parameters["positions"] = [0.25, 0.5, 0.75]
        annotation_ = self.plotter.annotate(self.annotation)
        for drawing in annotation_.drawings:
            self.assertEqual(color, drawing.get_color())

    def test_annotate_with_limits_sets_limits(self):
        self.annotation.parameters["positions"] = [0.5]
        self.annotation.parameters["limits"] = [0.25, 0.75]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertListEqual(
            annotation_.parameters["limits"],
            annotation_.drawings[0].get_ydata(),
        )


class TestHorizontalLine(unittest.TestCase):
    def setUp(self):
        self.annotation = annotation.HorizontalLine()
        self.plotter = aspecd.plotting.Plotter()

    def test_instantiate_class(self):
        pass

    def test_annotate_adds_line_to_plotter(self):
        self.annotation.parameters["positions"] = [0.5]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_adds_line_at_correct_position(self):
        self.annotation.parameters["positions"] = [0.5]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            annotation_.parameters["positions"][0],
            annotation_.drawings[0].get_ydata()[0],
        )

    def test_annotate_adds_lines_to_plotter(self):
        self.annotation.parameters["positions"] = [0.25, 0.5, 0.75]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            len(annotation_.parameters["positions"]),
            len(annotation_.drawings),
        )
        for drawing in annotation_.drawings:
            self.assertIn(drawing, self.plotter.ax.get_children())

    def test_annotate_adds_line_to_plotter_after_plotting(self):
        self.annotation.parameters["positions"] = [0.5]
        annotation_ = self.plotter.annotate(self.annotation)
        self.plotter.plot()
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_set_line_colour_from_dict(self):
        line_colour = "#cccccc"
        properties = {"color": line_colour}
        self.annotation.properties.from_dict(properties)
        self.assertEqual(line_colour, self.annotation.properties.color)

    def test_annotate_sets_correct_line_color(self):
        color = "#cccccc"
        properties = {"color": color}
        self.annotation.properties.from_dict(properties)
        self.plotter.plot()
        self.annotation.parameters["positions"] = [0.5]
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(color, annotation_.drawings[0].get_color())

    def test_annotate_sets_correct_line_color_for_each_line(self):
        color = "#cccccc"
        properties = {"color": color}
        self.annotation.properties.from_dict(properties)
        self.plotter.plot()
        self.annotation.parameters["positions"] = [0.25, 0.5, 0.75]
        annotation_ = self.plotter.annotate(self.annotation)
        for drawing in annotation_.drawings:
            self.assertEqual(color, drawing.get_color())

    def test_annotate_with_limits_sets_limits(self):
        self.annotation.parameters["positions"] = [0.5]
        self.annotation.parameters["limits"] = [0.25, 0.75]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertListEqual(
            annotation_.parameters["limits"],
            annotation_.drawings[0].get_xdata(),
        )


class TestText(unittest.TestCase):
    def setUp(self):
        self.annotation = annotation.Text()
        self.plotter = aspecd.plotting.Plotter()

    def test_instantiate_class(self):
        pass

    def test_annotate_adds_text_to_plotter(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_adds_text_at_correct_position(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertListEqual(
            annotation_.parameters["positions"][0],
            list(annotation_.drawings[0].get_position()),
        )

    def test_annotate_adds_correct_text(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        result = [
            item
            for item in self.plotter.ax.get_children()
            if item is annotation_.drawings[0]
        ][0]
        self.assertEqual(
            annotation_.parameters["texts"][0], result.get_text()
        )

    def test_annotate_adds_texts_to_plotter(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5], [0.7, 0.7]]
        self.annotation.parameters["texts"] = ["foo", "bar"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            len(annotation_.parameters["positions"]),
            len(annotation_.drawings),
        )
        for drawing in annotation_.drawings:
            self.assertIn(drawing, self.plotter.ax.get_children())

    def test_annotate_with_xpositions_and_ypositions(self):
        self.annotation.parameters["xpositions"] = [0.3]
        self.annotation.parameters["ypositions"] = [0.7]
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_xpositions_as_ndarray_and_ypositions(self):
        self.annotation.parameters["xpositions"] = np.asarray([0.3, 0.4])
        self.annotation.parameters["ypositions"] = [0.7, 0.7]
        self.annotation.parameters["texts"] = ["foo", "bar"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_xpositions_and_ypositions_zero(self):
        self.annotation.parameters["xpositions"] = [0]
        self.annotation.parameters["ypositions"] = [0]
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_xpositions_and_scalar_ypositions(self):
        self.annotation.parameters["xpositions"] = [0.3]
        self.annotation.parameters["ypositions"] = 0.7
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_xpositions_and_scalar_ypositions_zero(self):
        self.annotation.parameters["xpositions"] = [0.3]
        self.annotation.parameters["ypositions"] = 0
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_many_xpositions_and_one_ypositions(self):
        self.annotation.parameters["xpositions"] = [0.3, 0.5, 0.7]
        self.annotation.parameters["ypositions"] = [0.7]
        self.annotation.parameters["texts"] = ["foo", "bar", "baz"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_to_dict_contains_properties(self):
        dict_ = self.annotation.to_dict()
        self.assertIn("properties", dict_)


class TestVerticalSpan(unittest.TestCase):
    def setUp(self):
        self.annotation = annotation.VerticalSpan()
        self.plotter = aspecd.plotting.Plotter()

    def test_instantiate_class(self):
        pass

    def test_annotate_adds_span_to_plotter(self):
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_adds_span_at_correct_position(self):
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            annotation_.parameters["positions"][0][0],
            annotation_.drawings[0].get_corners()[0][0],
        )
        self.assertEqual(
            annotation_.parameters["positions"][0][1],
            annotation_.drawings[0].get_corners()[1][0],
        )

    def test_annotate_adds_spans_to_plotter(self):
        self.annotation.parameters["positions"] = [[0.1, 0.3], [0.5, 0.6]]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            len(annotation_.parameters["positions"]),
            len(annotation_.drawings),
        )
        for drawing in annotation_.drawings:
            self.assertIn(drawing, self.plotter.ax.get_children())

    def test_annotate_adds_span_to_plotter_after_plotting(self):
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        annotation_ = self.plotter.annotate(self.annotation)
        self.plotter.plot()
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_before_plotting_does_not_add_annotation_twice(self):
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        self.plotter.annotate(self.annotation)
        self.plotter.plot()
        self.assertEqual(1, len(self.plotter.annotations))

    def test_set_span_facecolour_from_dict(self):
        color = "#cccccc"
        properties = {"facecolor": color}
        self.annotation.properties.from_dict(properties)
        self.assertEqual(color, self.annotation.properties.facecolor)

    def test_annotate_sets_correct_facecolor(self):
        color = "#cccccc"
        properties = {"facecolor": color}
        self.annotation.properties.from_dict(properties)
        self.plotter.plot()
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            color,
            matplotlib.colors.to_hex(annotation_.drawings[0].get_facecolor()),
        )

    def test_annotate_sets_correct_facecolor_for_each_span(self):
        color = "#cccccc"
        properties = {"facecolor": color}
        self.annotation.properties.from_dict(properties)
        self.plotter.plot()
        self.annotation.parameters["positions"] = [[0.1, 0.3], [0.5, 0.6]]
        annotation_ = self.plotter.annotate(self.annotation)
        for drawing in annotation_.drawings:
            self.assertEqual(
                color, matplotlib.colors.to_hex(drawing.get_facecolor())
            )

    def test_annotate_with_limits_sets_limits(self):
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        self.annotation.parameters["limits"] = [0.25, 0.75]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertListEqual(
            annotation_.parameters["limits"],
            list(annotation_.drawings[0].get_corners()[:, 1][1:3]),
        )


class TestHorizontalSpan(unittest.TestCase):
    def setUp(self):
        self.annotation = annotation.HorizontalSpan()
        self.plotter = aspecd.plotting.Plotter()

    def test_instantiate_class(self):
        pass

    def test_annotate_adds_span_to_plotter(self):
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_adds_span_at_correct_position(self):
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            annotation_.parameters["positions"][0][0],
            annotation_.drawings[0].get_corners()[1][1],
        )
        self.assertEqual(
            annotation_.parameters["positions"][0][1],
            annotation_.drawings[0].get_corners()[2][1],
        )

    def test_annotate_adds_spans_to_plotter(self):
        self.annotation.parameters["positions"] = [[0.1, 0.3], [0.5, 0.6]]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            len(annotation_.parameters["positions"]),
            len(annotation_.drawings),
        )
        for drawing in annotation_.drawings:
            self.assertIn(drawing, self.plotter.ax.get_children())

    def test_annotate_adds_span_to_plotter_after_plotting(self):
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        annotation_ = self.plotter.annotate(self.annotation)
        self.plotter.plot()
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_before_plotting_does_not_add_annotation_twice(self):
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        self.plotter.annotate(self.annotation)
        self.plotter.plot()
        self.assertEqual(1, len(self.plotter.annotations))

    def test_set_span_facecolour_from_dict(self):
        color = "#cccccc"
        properties = {"facecolor": color}
        self.annotation.properties.from_dict(properties)
        self.assertEqual(color, self.annotation.properties.facecolor)

    def test_annotate_sets_correct_facecolor(self):
        color = "#cccccc"
        properties = {"facecolor": color}
        self.annotation.properties.from_dict(properties)
        self.plotter.plot()
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            color,
            matplotlib.colors.to_hex(annotation_.drawings[0].get_facecolor()),
        )

    def test_annotate_sets_correct_facecolor_for_each_span(self):
        color = "#cccccc"
        properties = {"facecolor": color}
        self.annotation.properties.from_dict(properties)
        self.plotter.plot()
        self.annotation.parameters["positions"] = [[0.1, 0.3], [0.5, 0.6]]
        annotation_ = self.plotter.annotate(self.annotation)
        for drawing in annotation_.drawings:
            self.assertEqual(
                color, matplotlib.colors.to_hex(drawing.get_facecolor())
            )

    def test_annotate_with_limits_sets_limits(self):
        self.annotation.parameters["positions"] = [[0.5, 0.6]]
        self.annotation.parameters["limits"] = [0.25, 0.75]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertListEqual(
            annotation_.parameters["limits"],
            list(annotation_.drawings[0].get_corners()[:, 0][0:2]),
        )


class TestTextWithLine(unittest.TestCase):
    def setUp(self):
        self.annotation = annotation.TextWithLine()
        self.plotter = aspecd.plotting.Plotter()

    def test_instantiate_class(self):
        pass

    def test_annotate_adds_text_to_plotter(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_adds_text_at_correct_position(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertListEqual(
            annotation_.parameters["positions"][0],
            list(annotation_.drawings[0].get_position()),
        )

    def test_annotate_adds_text_at_correct_offset(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["offsets"] = [[1, 1]]
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertListEqual(
            [1.5, 1.5],
            list(annotation_.drawings[0].get_position()),
        )

    def test_annotate_adds_correct_text(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        result = [
            item
            for item in self.plotter.ax.get_children()
            if item is annotation_.drawings[0]
        ][0]
        self.assertEqual(
            annotation_.parameters["texts"][0], result.get_text()
        )

    def test_annotate_sets_correct_horizontal_alignment(self):
        self.annotation.parameters["positions"] = [
            [0.5, 0.5],
            [0.5, 0.5],
            [0.5, 0.5],
        ]
        self.annotation.parameters["offsets"] = [[0, 1], [1, 1], [-1, 1]]
        self.annotation.parameters["texts"] = ["foo", "bar", "baz"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            "center", annotation_.drawings[0].get_horizontalalignment()
        )
        self.assertEqual(
            "left", annotation_.drawings[1].get_horizontalalignment()
        )
        self.assertEqual(
            "right", annotation_.drawings[2].get_horizontalalignment()
        )

    def test_annotate_sets_correct_rel_position(self):
        self.annotation.parameters["positions"] = [
            [0.5, 0.5],
            [0.5, 0.5],
            [0.5, 0.5],
            [0.5, 0.5],
            [0.5, 0.5],
            [0.5, 0.5],
        ]
        self.annotation.parameters["offsets"] = [
            [0, 1],
            [1, 1],
            [-1, 1],
            [0, -1],
            [1, -1],
            [-1, -1],
        ]
        self.annotation.parameters["texts"] = [
            "foo",
            "bar",
            "baz",
            "-foo",
            "-bar",
            "-baz",
        ]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        relpos = [
            [0.5, 0.5],
            [0, 0],
            [1, 0],
            [0.5, 0.5],
            [0, 1],
            [1, 1],
        ]
        for idx, annotation in enumerate(annotation_.drawings):
            self.assertListEqual(relpos[idx], annotation.arrowprops["relpos"])

    def test_annotate_sets_correct_connection_style(self):
        self.annotation.parameters["positions"] = [
            [0.5, 0.5],
            [0.5, 0.5],
            [0.5, 0.5],
        ]
        self.annotation.parameters["offsets"] = [[0, 1], [1, 1], [-1, 1]]
        self.annotation.parameters["texts"] = ["foo", "bar", "baz"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIsNone(
            annotation_.drawings[0].arrowprops["connectionstyle"]
        )
        self.assertIn(
            "angleA=-135",
            annotation_.drawings[1].arrowprops["connectionstyle"],
        )
        self.assertIn(
            "angleA=-45",
            annotation_.drawings[2].arrowprops["connectionstyle"],
        )

    def test_annotate_adds_texts_to_plotter(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5], [0.7, 0.7]]
        self.annotation.parameters["texts"] = ["foo", "bar"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            len(annotation_.parameters["positions"]),
            len(annotation_.drawings),
        )
        for drawing in annotation_.drawings:
            self.assertIn(drawing, self.plotter.ax.get_children())

    def test_annotate_with_xpositions_and_ypositions(self):
        self.annotation.parameters["xpositions"] = [0.3]
        self.annotation.parameters["ypositions"] = [0.7]
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_xpositions_as_ndarray_and_ypositions(self):
        self.annotation.parameters["xpositions"] = np.asarray([0.3, 0.5])
        self.annotation.parameters["ypositions"] = [0.7, 0.7]
        self.annotation.parameters["texts"] = ["foo", "bar"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_xpositions_and_scalar_ypositions(self):
        self.annotation.parameters["xpositions"] = [0.3]
        self.annotation.parameters["ypositions"] = 0.7
        self.annotation.parameters["texts"] = ["foo"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_many_xpositions_and_one_ypositions(self):
        self.annotation.parameters["xpositions"] = [0.3, 0.5, 0.7]
        self.annotation.parameters["ypositions"] = [0.7]
        self.annotation.parameters["texts"] = ["foo", "bar", "baz"]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_to_dict_contains_properties(self):
        dict_ = self.annotation.to_dict()
        self.assertIn("properties", dict_)


class TestMarker(unittest.TestCase):
    def setUp(self):
        self.annotation = annotation.Marker()
        self.plotter = aspecd.plotting.Plotter()

    def test_instantiate_class(self):
        pass

    def test_annotate_adds_marker_to_plotter(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_adds_marker_at_correct_position(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertListEqual(
            annotation_.parameters["positions"][0],
            list(annotation_.drawings[0].get_data()),
        )

    def test_annotate_adds_correct_marker(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        result = [
            item
            for item in self.plotter.ax.get_children()
            if item is annotation_.drawings[0]
        ][0]
        self.assertEqual(
            annotation_.parameters["marker"][0], result.get_marker()
        )

    def test_annotate_with_marker_keywords(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["marker"] = "octagon"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        result = [
            item
            for item in self.plotter.ax.get_children()
            if item is annotation_.drawings[0]
        ][0]
        self.assertEqual("8", result.get_marker())

    def test_annotate_with_mathtext(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["marker"] = "$\mathcal{A}$"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        result = [
            item
            for item in self.plotter.ax.get_children()
            if item is annotation_.drawings[0]
        ][0]
        self.assertEqual(
            annotation_.parameters["marker"], result.get_marker()
        )

    def test_annotate_adds_markers_to_plotter(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5], [0.7, 0.7]]
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            len(annotation_.parameters["positions"]),
            len(annotation_.drawings),
        )
        for drawing in annotation_.drawings:
            self.assertIn(drawing, self.plotter.ax.get_children())

    def test_annotate_with_multiple_markers_does_not_add_connecting_line(
        self,
    ):
        self.annotation.parameters["positions"] = [[0.5, 0.5], [0.7, 0.7]]
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        for drawing in annotation_.drawings:
            self.assertEqual("None", drawing.get_linestyle())

    def test_annotate_with_xpositions_and_ypositions(self):
        self.annotation.parameters["xpositions"] = [0.3]
        self.annotation.parameters["ypositions"] = [0.7]
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_xpositions_as_ndarray_and_ypositions(self):
        self.annotation.parameters["xpositions"] = np.asarray([0.3, 0.4])
        self.annotation.parameters["ypositions"] = [0.7, 0.7]
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_xpositions_and_ypositions_zero(self):
        self.annotation.parameters["xpositions"] = [0]
        self.annotation.parameters["ypositions"] = [0]
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_xpositions_and_scalar_ypositions(self):
        self.annotation.parameters["xpositions"] = [0.3]
        self.annotation.parameters["ypositions"] = 0.7
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_xpositions_and_scalar_ypositions_zero(self):
        self.annotation.parameters["xpositions"] = [0.3]
        self.annotation.parameters["ypositions"] = 0
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_annotate_with_many_xpositions_and_one_ypositions(self):
        self.annotation.parameters["xpositions"] = [0.3, 0.5, 0.7]
        self.annotation.parameters["ypositions"] = [0.7]
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())

    def test_to_dict_contains_properties(self):
        dict_ = self.annotation.to_dict()
        self.assertIn("properties", dict_)

    def test_annotate_sets_correct_facecolor_for_each_mark(self):
        color = "#cccccc"
        properties = {"facecolor": color}
        self.annotation.properties.from_dict(properties)
        self.plotter.plot()
        self.annotation.parameters["positions"] = [[0.1, 0.3], [0.5, 0.6]]
        annotation_ = self.plotter.annotate(self.annotation)
        for drawing in annotation_.drawings:
            self.assertEqual(
                color, matplotlib.colors.to_hex(drawing.get_markerfacecolor())
            )

    def test_annotate_with_scalar_yoffset(self):
        self.annotation.parameters["positions"] = [[0.5, 0.5]]
        self.annotation.parameters["yoffset"] = 0.1
        self.annotation.parameters["marker"] = "o"
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            annotation_.parameters["positions"][0][1]
            + self.annotation.parameters["yoffset"],
            annotation_.drawings[0].get_data()[1],
        )


class TestFillBetween(unittest.TestCase):
    def setUp(self):
        self.annotation = annotation.FillBetween()
        self.plotter = aspecd.plotting.Plotter()
        self.dataset = aspecd.dataset.CalculatedDataset()
        self.dataset.data.data = np.ones(101)

    def test_instantiate_class(self):
        pass

    def test_annotate_adds_annotation_to_plotter(self):
        self.annotation.parameters["data"] = self.dataset
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertIn(annotation_.drawings[0], self.plotter.ax.get_children())
        self.assertIsInstance(
            annotation_.drawings[0], matplotlib.collections.PolyCollection
        )

    def test_annotate_adds_annotations_to_plotter(self):
        self.annotation.parameters["data"] = [self.dataset, self.dataset]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertEqual(
            len(annotation_.parameters["data"]),
            len(annotation_.drawings),
        )
        for drawing in annotation_.drawings:
            self.assertIn(drawing, self.plotter.ax.get_children())

    def test_to_dict_contains_properties(self):
        dict_ = self.annotation.to_dict()
        self.assertIn("properties", dict_)

    def test_annotate_sets_correct_facecolor_for_each_annotation(self):
        color = "#cccccc"
        properties = {"facecolor": color}
        self.annotation.properties.from_dict(properties)
        self.plotter.plot()
        self.annotation.parameters["data"] = [self.dataset, self.dataset]
        annotation_ = self.plotter.annotate(self.annotation)
        for drawing in annotation_.drawings:
            self.assertEqual(
                color, matplotlib.colors.to_hex(drawing.get_facecolor())
            )

    def test_annotate_with_second_as_scalar(self):
        self.annotation.parameters["data"] = self.dataset
        self.annotation.parameters["second"] = 0.1
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertAlmostEqual(
            0.1,
            annotation_.drawings[0].get_paths()[0].get_extents().bounds[1],
        )

    def test_annotate_with_second_as_dataset(self):
        self.annotation.parameters["data"] = self.dataset
        second = aspecd.dataset.CalculatedDataset()
        second.data.data = np.ones(101) * 0.1
        self.annotation.parameters["second"] = second
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertAlmostEqual(
            0.1,
            annotation_.drawings[0].get_paths()[0].get_extents().bounds[1],
        )

    def test_annotate_with_seconds_as_list_of_datasets(self):
        self.annotation.parameters["data"] = self.dataset
        second = aspecd.dataset.CalculatedDataset()
        second.data.data = np.ones(101) * 0.1
        self.annotation.parameters["second"] = [second, second]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertAlmostEqual(
            0.1,
            annotation_.drawings[0].get_paths()[0].get_extents().bounds[1],
        )

    def test_annotate_with_seconds_as_list_of_scalars(self):
        self.annotation.parameters["data"] = self.dataset
        self.annotation.parameters["second"] = [0.2, 0.5]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertAlmostEqual(
            0.2,
            annotation_.drawings[0].get_paths()[0].get_extents().bounds[1],
        )

    def test_annotate_with_seconds_as_mixed_list(self):
        self.annotation.parameters["data"] = [self.dataset, self.dataset]
        second = aspecd.dataset.CalculatedDataset()
        second.data.data = np.ones(101) * 0.1
        self.annotation.parameters["second"] = [0.2, second]
        self.plotter.plot()
        annotation_ = self.plotter.annotate(self.annotation)
        self.assertAlmostEqual(
            0.2,
            annotation_.drawings[0].get_paths()[0].get_extents().bounds[1],
        )
        self.assertAlmostEqual(
            0.1,
            annotation_.drawings[1].get_paths()[0].get_extents().bounds[1],
        )



"""General purpose functions and classes used in other modules.

To avoid circular dependencies, this module does *not* depend on any other
modules of the ASpecD package, but it can be imported into every other module.

"""

import collections
import contextlib
import datetime
import hashlib
import importlib
import inspect
import logging
import os
import pkgutil
import re

import numpy as np
import oyaml as yaml
import pkg_resources

import aspecd.exceptions


def full_class_name(object_):
    """
    Return full class name of an object including packages and modules.

    Parameters
    ----------
    object_ : `object`
        object the class name should be inferred for

    Returns
    -------
    class_name : :class:`str`
        string with full class name of object

    """
    class_name = "".join(
        [object_.__class__.__module__, ".", object_.__class__.__name__]
    )
    return class_name


def object_from_class_name(full_class_name_string):
    """
    Create object from full class name.

    To obtain the full class name of an object, you might want to use the
    function :func:`full_class_name`

    Parameters
    ----------
    full_class_name_string : :class:`str`
        string with full class name of an object that shall be instantiated

    Returns
    -------
    object_ : `object`
        object instantiated from the class given in `full_class_name_string`

    """
    class_name_parts = full_class_name_string.split(".")
    class_name = class_name_parts[-1]
    module_name = ".".join(class_name_parts[0:-1])
    module = importlib.import_module(module_name)
    object_ = getattr(module, class_name)()
    return object_


class ToDictMixin:
    """Mixin class for returning all public attributes as dict.

    Sometimes there is the need to either exclude public attributes (in case
    of infinite loops created by trying to apply ``to_dict`` in this case)
    or to add (public) attributes, particularly those used by getters and
    setters that are otherwise not included.

    To do so, there are two non_public attributes of this class each class
    inheriting from it will be able to set as well:

    * :attr:`_exclude_from_to_dict`
    * :attr:`_include_in_to_dict`

    The names should be rather telling. For details, see below.

    Attributes
    ----------
    __odict__ : :class:`collections.OrderedDict`
        Dictionary of attributes preserving the order of their definition

    _exclude_from_to_dict : :class:`list`
        Names of (public) attributes to exclude from dictionary

        Usually, the reason to exclude public attributes from being added to
        the dictionary is to avoid infinite loops, as sometimes an object
        may contain a reference to another object that in turn references
        back.

    _include_in_to_dict : :class:`list`
        Names of (public) attributes to include into dictionary

        Usual reasons for actively including (public) attributes into the
        dictionary are those attributes accessed by getters and setters and
        hence not automatically included in the list otherwise.


    .. versionchanged: 0.6.3
        Added superclass call to preserve mro in dependent subclasses

    """

    def __init__(self):
        super().__init__()
        if "__odict__" not in self.__dict__:
            self.__odict__ = collections.OrderedDict()
        self._exclude_from_to_dict = []
        self._include_in_to_dict = []

    def __setattr__(self, attribute, value):
        """
        Add attributes to :attr:`__odict__` to preserve order of definition.

        Parameters
        ----------
        attribute : :class:`str`
            Name of attribute
        value : :class:`str`
            Value of attribute

        """
        if "__odict__" not in self.__dict__:
            super().__setattr__("__odict__", collections.OrderedDict())
        self.__odict__[attribute] = value
        super().__setattr__(attribute, value)

    def to_dict(self, remove_empty=False):
        """
        Create dictionary containing public attributes of an object.

        Parameters
        ----------
        remove_empty : :class:`bool`
            Whether to remove keys with empty values

            Default: False

        Returns
        -------
        public_attributes : :class:`collections.OrderedDict`
            Ordered dictionary containing the public attributes of the object

            The order of attribute definition is preserved


        .. versionchanged:: 0.6
            New parameter `remove_empty`

        .. versionchanged:: 0.9
            Settings for properties to exclude and include are not traversed

        .. versionchanged:: 0.9.1
            Dictionaries get copied before traversing, as otherwise,
            the special variables ``__dict__`` and  ``__0dict__`` are
            modified, what may result in strange behaviour.

        .. versionchanged:: 0.9.2
            Dictionaries do not get copied by default, but there is a
            private method that can be overridden in derived classes to
            copy the dictionary.

        """
        if hasattr(self, "__odict__"):
            result = self._traverse_dict(
                self._clean_dict(
                    self._return_copy(self.__odict__), traversing=False
                )
            )
        else:
            result = self._traverse_dict(
                self._clean_dict(
                    self._return_copy(self.__dict__), traversing=False
                )
            )
        if remove_empty:
            result = remove_empty_values_from_dict(result)
        return result

    @staticmethod
    def _return_copy(dictionary):
        """
        Return (a copy of) the dictionary.

        Some derived classes need to operate on a (deep)copy of the
        dictionary, while others don't. Generally, (deep)copying slows down
        things, hence should be avoided wherever possible.

        Parameters
        ----------
        dictionary : :class:`dict`
            The dictionary to be copied

        Returns
        -------
        dictionary : class:`dict`
            The (copy of the) dictionary

        """
        return dictionary

    def _clean_dict(self, dictionary, traversing=True):
        to_remove = []
        for key in dictionary:
            if (
                (
                    str(key).startswith("_")
                    and key not in self._include_in_to_dict
                )
                or str(key) in self._exclude_from_to_dict
                and not traversing
            ):
                to_remove.append(key)
        for key in to_remove:
            dictionary.pop(key, None)
        if not traversing:
            for key in self._include_in_to_dict:
                dictionary[key] = getattr(self, key)
        return dictionary

    def _traverse_dict(self, instance_dict):
        output = collections.OrderedDict()
        for key, value in instance_dict.items():
            output[key] = self._traverse(key, value)
        return output

    def _traverse(self, key, value):
        if isinstance(value, ToDictMixin):
            result = value.to_dict()
        elif isinstance(value, (dict, collections.OrderedDict)):
            result = self._traverse_dict(self._clean_dict(value))
        elif hasattr(value, "__odict__"):
            result = self._traverse_dict(self._clean_dict(value.__odict__))
        elif hasattr(value, "__dict__"):
            result = self._traverse_dict(self._clean_dict(value.__dict__))
        elif isinstance(value, list):
            result = [self._traverse(key, i) for i in value]
        elif isinstance(
            value, (datetime.datetime, datetime.date, datetime.time)
        ):
            result = str(value)
        else:
            result = value
        return result


def get_aspecd_version():
    """
    Get version of ASpecD package.

    The function directly reads the contents of the file VERSION in the root
    directory of the package installation, not relying on introspection.

    Returns
    -------
    version : :class:`str`
        Version number as string

    """
    version_file_path = os.path.join(
        os.path.dirname(__file__), "..", "VERSION"
    )
    if os.path.exists(version_file_path):
        with open(version_file_path, encoding="utf8") as version_file:
            version = version_file.read().strip()
    else:
        version = package_version("aspecd")
    return version


def package_version(name=""):
    """
    Get version of arbitrary package.

    The function relies on introspection using :mod:`pkg_resources`.

    Parameters
    ----------
    name : :class:`str`
        Name of package the version should be obtained for

    Returns
    -------
    version : :class:`str`
        Version number as string

    """
    version = None
    if name:
        try:
            version = pkg_resources.get_distribution(name).version
        except pkg_resources.DistributionNotFound:
            pass
    return version


def package_name(obj=None):
    """
    Get name of package an object resides in.

    Parameters
    ----------
    obj : :class:`object`
        Object the package it resides in should be returned.

        If no object is given, the name of the package this function is
        defined in ("aspecd") will be returned.

    Returns
    -------
    package_name : :class:`str`
        Name of the package

    """
    if not obj:
        obj = package_name
    module = inspect.getmodule(obj)
    return module.__package__


def config_dir():
    """
    Get config directory for per-user configurations.

    Configuration on a per-user level should be stored within a directory
    (only) readable by the currently logged-in user.

    Returns
    -------
    config_dir : :class:`str`
        Path to config directory, usually in the user's directory

    """
    config_dir_ = os.environ.get(
        "XDG_CONFIG_HOME", os.path.join(os.path.expanduser("~"), ".config")
    )
    return config_dir_


class Yaml:
    """
    Handle reading from and writing to YAML files.

    YAML file contents are read into an ordered dict, making use of the
    oyaml package. This preserves the order of the entries of any dict.

    .. note::
        The PyYAML package cannot handle NumPy arrays by default. Hence,
        there are two methods in this class for serialising and
        deserialising NumPy arrays:

        * :meth:`aspecd.utils.Yaml.serialise_numpy_arrays` and
        * :meth:`aspecd.utils.Yaml.deserialise_numpy_arrays`.

        Have a look at their documentation for details of the implementation
        and how to use them. In particular, larger NumPy arrays are saved to
        files using the NumPy binary format.

    Attributes
    ----------
    dict : :class:`collections.OrderedDict`
        Contents read from/written to a YAML file

    numpy_array_size_threshold : :class:`int`
        Maximum size of NumPy array that is converted into a list using
        :meth:`aspecd.utils.Yaml.serialise_numpy_arrays`.

        Larger NumPy arrays are saved to a file in NumPy format using
        :func:`numpy.save`.

        Default: 100 -- *i.e.*, arrays with >100 elements are stored in files.

    numpy_array_to_list : class:`bool`
        Whether to convert (small) NumPy arrays to a list or to a dict

        Default: False

    binary_files : :class:`list`
        List of names of the binary files containing large arrays

    binary_directory : :class:`str`
        Directory the binary files should be stored in

        Default: ''

    loader : :class:`yaml.loader`
        Type of loader used for loading the YAML file

        The type of loader used is crucial for the safety of your
        application. See the documentation of the PyYAML package for details.

        Default: :class:`yaml.SafeLoader`

    dumper : :class:`yaml.Dumper`
        Type of dumper used for loading the YAML file

        The type of dumper used is should be compatible to the type of loader.

        Default: :class:`yaml.SafeDumper`

    Raises
    ------
    aspecd.utils.MissingFilenameError
        Raised if no filename is given to read from/write to


    .. versionchanged:: 0.4
        Added :attr:`numpy_array_to_list`

    .. versionchanged:: 0.5
        Added :attr:`dumper` and set dumper to SafeDumper

    """

    def __init__(self):
        self.binary_files = []
        self.binary_directory = ""
        self.dict = collections.OrderedDict()
        self.numpy_array_size_threshold = 100
        self.numpy_array_to_list = False
        self.loader = yaml.SafeLoader
        self.dumper = yaml.SafeDumper
        self.loader.add_implicit_resolver(
            "tag:yaml.org,2002:float",
            re.compile(
                """^(?:
             [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
            |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
            |\\.[0-9_]+(?:[eE][-+][0-9]+)?
            |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
            |[-+]?\\.(?:inf|Inf|INF)
            |\\.(?:nan|NaN|NAN))$""",
                re.X,
            ),
            list("-+0123456789."),
        )
        self.dumper.add_representer(np.float64, self._numpy_float_representer)
        self.dumper.add_representer(np.int64, self._numpy_int_representer)

    def read_from(self, filename=""):
        """
        Read from YAML file.

        Parameters
        ----------
        filename : :class:`str`
            Name of the YAML file to read from.

        Raises
        ------
        aspecd.processing.MissingFilenameError
            Raised if no filename is given to read from.

        """
        if not filename:
            raise aspecd.exceptions.MissingFilenameError
        with open(filename, "r", encoding="utf8") as file:
            self.dict = yaml.load(file, Loader=self.loader)

    def write_to(self, filename=""):
        """
        Write to YAML file.

        Parameters
        ----------
        filename : :class:`str`
            Name of the YAML file to write to.

        Raises
        ------
        aspecd.processing.MissingFilenameError
            Raised if no filename is given to write to.

        """
        if not filename:
            raise aspecd.exceptions.MissingFilenameError
        with open(filename, "w", encoding="utf8") as file:
            yaml.dump(self.dict, file, Dumper=self.dumper)

    def read_stream(self, stream=None):
        """
        Read from stream.

        Parameters
        ----------
        stream : :class:`bytes`
            binary stream to read from

        """
        self.dict = yaml.load(stream, Loader=self.loader)

    def write_stream(self):
        """
        Write to from stream.

        Returns
        -------
        stream : :class:`str`
            string representation of YAML file

        """
        return yaml.dump(self.dict, Dumper=self.dumper)

    def serialise_numpy_arrays(self):
        """
        Serialise numpy arrays in a simple form, using a dict.

        Background: The PyYAML package cannot easily handle NumPy arrays out
        of the box. However, datasets and alike should sometimes be
        serialised in form of YAML files. Hence the need for a simple
        method of serialising NumPy arrays. Furthermore, larger NumPy
        arrays should not be serialised in text form in YAML directly,
        but rather be stored in binary form, probably in a separate file.

        The reason for saving (larger) NumPy arrays in binary rather than
        text form is twofold: The size of a binary file is much smaller,
        and the original precision is retained. Those interested in the
        representation of floating-point values in computers should consult:

        * David Goldberg. What every computer scientist should know about
          floating-point arithmetic. *ACM Comput. Surv.* **23** (1991):5--48.
          DOI:`10.1145/103162.103163 <https://doi.org/10.1145/103162.103163>`_

        As binary format, the NumPy format (see :mod:`numpy.lib.format`)
        gets used.

        As filename, the SHA256 hash of the array will be used. Thus,
        the names are unique with respect to the content. In the event of
        having several identical arrays within one dict that gets
        serialised, this should not be a problem, as the hashes should be
        reasonably unique. *I.e.*, identical files should have identical
        content. Thus, having several identical arrays will lead to less
        files written, eventually saving space and overall file size.

        """
        self._traverse_serialise_numpy_arrays(dict_=self.dict)

    def _traverse_serialise_numpy_arrays(self, dict_=None):  # noqa: MC0001
        for key in dict_.keys():
            if isinstance(dict_[key], list):
                for element in dict_[key]:
                    if isinstance(element, (dict, collections.OrderedDict)):
                        self._traverse_serialise_numpy_arrays(dict_=element)
            elif isinstance(dict_[key], np.ndarray):
                if dict_[key].size > self.numpy_array_size_threshold:
                    self._create_binary_directory()
                    try:
                        filename = (
                            hashlib.sha256(dict_[key]).hexdigest() + ".npy"
                        )
                    except ValueError:
                        filename = (
                            hashlib.sha256(dict_[key].copy()).hexdigest()
                            + ".npy"
                        )
                    np.save(
                        os.path.join(self.binary_directory, filename),
                        dict_[key],
                        allow_pickle=False,
                    )
                    dict_[key] = {
                        "type": "numpy.ndarray",
                        "dtype": str(dict_[key].dtype),
                        "file": filename,
                    }
                    self.binary_files.append(filename)
                elif self.numpy_array_to_list:
                    dict_[key] = dict_[key].tolist()
                else:
                    dict_[key] = {
                        "type": "numpy.ndarray",
                        "dtype": str(dict_[key].dtype),
                        "array": dict_[key].tolist(),
                    }
            elif isinstance(dict_[key], (dict, collections.OrderedDict)):
                self._traverse_serialise_numpy_arrays(dict_=dict_[key])
            # make list of binary_files unique
            self.binary_files = list(set(self.binary_files))

    def serialize_numpy_arrays(self):
        """
        Serialise numpy arrays for our AE speaking friends.

        See :meth:`aspecd.utils.Yaml.serialise_numpy_arrays` for details.
        """
        self.serialise_numpy_arrays()

    def deserialise_numpy_arrays(self):
        """
        Deserialise specially crafted dicts into NumPy arrays.

        For reasons given in the documentation of
        :meth:`aspecd.utils.Yaml.serialise_numpy_arrays`, NumPy arrays are
        handled separately and converted into dicts with some special
        fields. This method deserialises them back into NumPy arrays,
        as they were originally.

        """
        self._traverse_deserialise_numpy_arrays(dict_=self.dict)

    def _traverse_deserialise_numpy_arrays(self, dict_=None):
        for key in dict_.keys():
            if isinstance(dict_[key], list):
                for element in dict_[key]:
                    if type(element) in [dict, collections.OrderedDict]:
                        self._traverse_deserialise_numpy_arrays(dict_=element)
            elif isinstance(dict_[key], (dict, collections.OrderedDict)):
                if (
                    "type" in dict_[key].keys()
                    and dict_[key]["type"] == "numpy.ndarray"
                ):
                    if "file" in dict_[key].keys():
                        self._create_binary_directory()
                        dict_[key] = np.load(
                            os.path.join(
                                self.binary_directory, dict_[key]["file"]
                            )
                        )
                    else:
                        dict_[key] = np.asarray(dict_[key]["array"])
                else:
                    self._traverse_deserialise_numpy_arrays(dict_=dict_[key])

    def deserialize_numpy_arrays(self):
        """
        Deserialise numpy arrays for our AE speaking friends.

        See :meth:`aspecd.utils.Yaml.deserialise_numpy_arrays` for details.
        """
        self.deserialise_numpy_arrays()

    def _create_binary_directory(self):
        if self.binary_directory and not os.path.exists(
            self.binary_directory
        ):
            os.mkdir(self.binary_directory)

    @staticmethod
    def _numpy_float_representer(dumper, data):
        return dumper.represent_float(float(data))

    @staticmethod
    def _numpy_int_representer(dumper, data):
        return dumper.represent_int(int(data))


def replace_value_in_dict(replacement=None, target=None):  # noqa: MC0001
    """
    Replace value for given key in a dictionary, traversing recursively.

    The key in ``replacement`` needs to correspond to the value in
    ``target_dict`` that should be replaced. Keys in the dict ``replacement``
    that have no corresponding value in ``target_dict`` are silently ignored.

    Parameters
    ----------
    replacement : :class:`dict`
        dict containing key corresponding to the value in ``target_dict``
        that should be replaced by the associated value
    target : :class:`dict`
        dict containing the key whose value should be replaced by the value
        of the key in ``replacement`` named identical to the value

    Returns
    -------
    target : :class:`dict`
        dict containing the key whose value has been replaced by the value
        of the corresponding key in ``replacement``

    """
    for key, value in target.items():
        if isinstance(value, np.ndarray):
            continue
        if isinstance(target[key], dict):
            target[key] = replace_value_in_dict(replacement, target[key])
        elif isinstance(target[key], list):
            if target[key]:
                for list_index, list_element in enumerate(target[key]):
                    if isinstance(list_element, dict):
                        target[key][list_index] = replace_value_in_dict(
                            replacement, list_element
                        )
                    elif (
                        list_element.__hash__ and list_element in replacement
                    ):
                        target[key][list_index] = replacement[list_element]
                    else:
                        target[key][list_index] = list_element
            elif key in replacement:
                target[key] = replacement[key]
        elif value in replacement:
            target[key] = replacement[value]
    return target


def copy_values_between_dicts(source=None, target=None):
    """
    Copy values between two dicts in case of identical keys.

    Each value in ``source`` is copied to ``target`` in case of matching
    keys in a recursive manner. Non-matching keys in ``source`` are
    silently ignored.

    Both, source and target are parallel recursively traversed in case of
    identical overall structure.

    Parameters
    ----------
    source : :class:`dict`
        Dictionary the values should be copied from in case of matching keys
    target : :class:`dict`
        Dictionary the values should be copied to in case of matching keys

    Returns
    -------
    target : :class:`dict`
        Dictionary the values have been copied to in case of matching keys

    """
    for key in target:
        if isinstance(target[key], dict):
            if key in source and isinstance(source[key], dict):
                target[key] = copy_values_between_dicts(
                    source[key], target[key]
                )
            else:
                target[key] = copy_values_between_dicts(source, target[key])
        elif key in source:
            target[key] = source[key]
    return target


def copy_keys_between_dicts(source=None, target=None):
    """
    Copy keys between two dicts.

    Each key in ``source`` is copied to ``target`` in a recursive manner.
    If the key in ``source`` is a dict and exists in ``target``, the two
    dicts will be joined, not loosing keys in ``target``.

    If, however, the key in ``source`` is *not* a dict, but the
    corresponding key in ``target`` is a dict, the corresponding value in
    ``target`` will be replaced with that from ``source``.

    Parameters
    ----------
    source : :class:`dict`
        Dictionary the keys should be copied from
    target : :class:`dict`
        Dictionary the keys should be copied to

    Returns
    -------
    target : :class:`dict`
        Dictionary the keys have been copied to

    """
    for key in source:
        if (
            key in target
            and isinstance(target[key], dict)
            and isinstance(source[key], dict)
        ):
            target[key] = copy_keys_between_dicts(source[key], target[key])
        else:
            target[key] = source[key]
    return target


def remove_empty_values_from_dict(dict_):
    """
    Remove empty values and their keys from dict recursively.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary the keys with empty values should be removed from

    Returns
    -------
    dict_ : :class:`dict`
        Dictionary the keys with empty values have been removed from


    .. versionadded:: 0.2.1

    """
    if isinstance(dict_, dict):
        dict_ = dict(
            (key, remove_empty_values_from_dict(value))
            for key, value in dict_.items()
            if value and remove_empty_values_from_dict(value)
        )
    return dict_


def convert_keys_to_variable_names(dict_):
    """
    Change keys in dict to comply with PEP8 for variable names.

    Keys are converted to all lower case and spaces replaced with underscores.

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary the keys should be renamed in

    Returns
    -------
    new_dict : :class:`dict`
        Dictionary the keys have been renamed in


    .. versionadded:: 0.2.1

    """
    new_dict = {}
    for key, value in dict_.items():
        new_key = key.replace(" ", "_").lower()
        if isinstance(value, dict):
            new_dict[new_key] = convert_keys_to_variable_names(dict_[key])
        else:
            new_dict[new_key] = dict_[key]
    return new_dict


def all_equal(list_=None):
    """
    Check whether all elements of a list are equal.

    Parameters
    ----------
    list_ : :class:`list`
        List whose elements should be checked for being equal

    Returns
    -------
    result : :class:`bool`

    """
    return list_.count(list_[0]) == len(list_)


class Properties(ToDictMixin):
    """
    General properties class allowing to set properties from dict.

    Properties classes often need to set their properties from dicts,
    and additionally, they should be able to convert themselves into dicts
    for persistence.
    """

    def __init__(self):
        super().__init__()
        self._exclude = []

    def from_dict(self, dict_=None):
        """
        Set attributes from dictionary.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary containing information of a task.

        Raises
        ------
        aspecd.plotting.MissingDictError
            Raised if no dict is provided.

        """
        if not dict_:
            raise aspecd.exceptions.MissingDictError
        for key, value in dict_.items():
            if hasattr(self, key):
                if hasattr(getattr(self, key), "from_dict"):
                    getattr(self, key).from_dict(value)
                elif isinstance(getattr(self, key), list):
                    if isinstance(value, list):
                        for element in value:
                            getattr(self, key).append(element)
                    else:
                        getattr(self, key).append(value)
                else:
                    setattr(self, key, value)

    def get_properties(self):
        """
        Return (public) properties, i.e. attributes that are *not* methods.

        Returns
        -------
        properties : :class:`list`
            public properties

        """
        props = []
        for prop in list(self.__dict__["__odict__"].keys()):
            if not str(prop).startswith("_") and prop not in self._exclude:
                props.append(prop)
        return props


def basename(filename):
    """
    Return basename of given filename.

    Parameters
    ----------
    filename : :class:`str`
        Name of the file (may contain absolute path and extension) the
        basename should be returned for.

    Returns
    -------
    basename : :class:`str`
        Basename corresponding to the filename provided as input.

    """
    return os.path.splitext(os.path.split(filename)[-1])[0]


def path(filename):
    """
    Return path of given filename, with trailing separator.

    Parameters
    ----------
    filename : :class:`str`
        Name of the file (may contain absolute path and extension) the path
        should be returned for.


    Returns
    -------
    path : :class:`str`
        path corresponding to the filename provided as input.

    """
    return os.path.split(filename)[:-1][0] + os.sep


def not_zero(value):
    """
    Return a value that is not zero to prevent DivisionByZero errors.

    Dividing by zero results in NaN values and often hinders evaluating
    mathematical models. A solution adopted from the lmfit Python package
    (https://doi.org/10.5281/zenodo.598352) returns a value equivalent to
    the resolution of a numpy float.

    .. note::

        If you use this function excessively within a module, mostly within
        rather complicated mathematical equations, it might be a good idea
        to import this function explicitly, to shorten the code, such as:
        ``from aspecd.utils import not_zero``. As usual, readability is king.

    Parameters
    ----------
    value : :class:`float`
        Value that can become (too close to) zero to trigger NaN values

    Returns
    -------
    value : :class:`float`
        Value guaranteed not to be zero


    .. versionadded:: 0.3

    """
    return np.copysign(
        max(abs(value), np.finfo(np.float64).resolution), value
    )


def isiterable(variable):
    """
    Check whether the given variable is iterable.

    Lists, tuples, NumPy arrays, but strings as well are iterable. Integers,
    however, are not.

    Parameters
    ----------
    variable :
        variable to check for being iterable

    Returns
    -------
    answer : :class:`bool`
        Whether the given variable is iterable

    """
    try:
        iter(variable)
        answer = True
    except TypeError:
        answer = False
    return answer


def get_package_data(name="", directory=""):
    """
    Obtain contents from a non-code file ("package data").

    A note on obtaining data from the distributed package: Rather than
    manually playing around with paths relative to the package root
    directory, contents of non-code files need to be obtained in a way that
    works with different kinds of package installation.

    Note that in Python, only files within the package, *i.e.* within the
    directory where all the modules are located, can be accessed, not files
    that reside on the root directory of the package.

    .. note::

        In case you would want to get package data from a package different
        than aspecd, you can prefix the name of the file to retrieve with
        the package name, using the '@' as a separator.

        Suppose you would want to retrieve the file ``__main__.py`` file
        from the "pip" package (why should you?):

        .. code-block::

            get_package_data('pip@__main__.py', directory='')

        This would return the contents of this file.


    Parameters
    ----------
    name : :class:`str`
        Name of the file whose contents should be accessed.

        In case the file should be retrieved from a different package,
        the package name can be prefixed, using '@' as a separator.

    directory : :class:`str`
        Directory within the package where the files are located.

        Default: ''

    Returns
    -------
    contents : :class:`str`
        String containing the contents of the non-code file.


    .. versionadded:: 0.5

    """
    if not name:
        raise ValueError("No filename given.")
    package = __package__
    if "@" in name:
        package, name = name.split("@")
    contents = pkgutil.get_data(package, "/".join([directory, name])).decode()
    return contents


# noinspection PyShadowingNames
@contextlib.contextmanager
def change_working_dir(path=""):  # pylint: disable=redefined-outer-name
    """
    Context manager for temporarily changing the working directory.

    Sometimes it is necessary to temporarily change the working directory,
    but one would like to ensure that the directory is reverted even in case
    an exception is raised.

    Due to its nature as a context manager, this function can be used with a
    ``with`` statement. See below for an example.


    Parameters
    ----------
    path : :class:`str`
        Path the current working directory should be changed to.


    Examples
    --------
    To temporarily change the working directory:

    .. code-block::

        with change_working_dir(os.path.join('some', 'path')):
            # Do something that may raise an exception

    This can come in quite handy in case of tests.


    .. versionadded:: 0.6

    """
    oldpwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)


def get_logger(name=""):
    """
    Get logger object for a given module.

    Logging in libraries is slightly different from standard logging with
    respect to the handler attached to the logger. The general advice from
    the `Python Logging HOWTO <https://docs.python.org/3/howto/logging.html
    #configuring-logging-for-a-library>`_ is to explicitly add the
    :class:`logging.Nullhandler` as a handler.

    Additionally, if you want to add loggers for a library that inherits
    from/builds upon a framework, in this particular case a library/package
    built atop the ASpecD framework, you want to have loggers being children
    of the framework logger in order to have the framework catch the log
    messages your library/package creates.

    Why does this matter, particularly for the ASpecD framework? If you want
    to have your log messages in a package based on the ASpecD framework appear
    when using :doc:`recipe-driven data analysis </recipes>`, you need to
    have your package loggers to be in the hierarchy below the root logger
    of the ASpecD framework. For convenience and in order not to make any
    informed guesses on how the ASpecD framework root logger is named,
    simply use this function to create the loggers in the modules of your
    package.

    Parameters
    ----------
    name : :class:`str`
        Name of the module to get the logger for.

        Usually, this will be set to ``__name__``, as this returns the
        current module (including the package name and separated with a
        ``.`` if present).

    Returns
    -------
    logger : :class:`logging.Logger`
        Logger object for the module

        The logger will have a :class:`logging.NullHandler` handler
        attached, in line with the advice from the `Python Logging HOWTO
        <https://docs.python.org/3/howto/logging.html#configuring-logging-for
        -a-library>`_.


    Examples
    --------
    To add a logger to a module in your library/package based on the ASpecD
    framework, add something like this to the top of your module:

    .. code-block::

        import aspecd.utils


        logger = aspecd.utils.get_logger(__name__)

    The important aspect here is to use ``__name__`` as the name of the
    logger. The reason is simple: ``__name__`` gets automatically expanded
    to the name of the current module, with the name of all parent
    modules/the package prefixed, using dot notation. The resulting logger
    will be situated in the hierarchy below the ``aspecd`` package logger.
    Suppose you have added the above lines to the module
    ``mypackage.processing`` in the ``processing`` module of your package
    ``mypackage``. This will result in a logger ``aspecd.mypackage.processing``.


    .. versionadded:: 0.9

    """
    if not name:
        name = package_name()
    else:
        name = ".".join([package_name(), name])
    logger = logging.getLogger(name=name)
    logger.addHandler(logging.NullHandler())
    return logger



"""
Numerical models

Models are defined by (constant) parameters and variables the model is
evaluated for. The variables can be thought of as the axes values of the
resulting (calculated) dataset.

As a simple example, consider a polynomial defined by its (constant)
coefficients. The model will evaluate the polynomial for the values,
and the result will be a :obj:`aspecd.dataset.CalculatedDataset` object
containing the values of the evaluated model in its data, and the
variables as its axes values.

Models can be seen as abstraction to simulations in some regard. In this
respect, they will play a central role in conjunction with fitting
models to data by adjusting their respective parameters, a quite general
approach in science and particularly in spectroscopy.


A bit of terminology
====================

parameters :
    constant parameters (sometimes termed coefficients) characterising the model

    Example: In case of a polynomial, the coefficients would be the
    parameters of the model.

variables :
    values to evaluate the model for

    Example: In case of a polynomial, the *x* values the model is
    evaluated for would be the variables, with the *y* values being the
    corresponding depending values dictated by the model and its parameters.


Models provided within this module
==================================

Besides providing the basis for models for the ASpecD framework, this module
comes with a (growing) number of general-purpose models useful for basically
all kinds of spectroscopic data.

Here is a list as a first overview. For details, see the detailed
documentation of each of the classes, readily accessible by the link.


Primitive models
----------------

Primitive models are mainly used to create test datasets that can be
operated on afterwards. The particular strength and beauty of wrapping
essential one-liners of code with a full-fledged model class is twofold:
These classes return ASpecD datasets, and you can work completely in context
of recipe-driven data analysis, requiring no actual programming skills.

If nothing else, these primitive models can serve as a way to create
datasets with fixed data dimensions. Those datasets may be used as templates
for more advanced models, by using the :meth:`aspecd.model.Model.from_dataset`
method.

Having that said, here you go with a list of primitive models:


* :class:`aspecd.model.Zeros`

  Dataset consisting entirely of zeros (in N dimensions)

* :class:`aspecd.model.Ones`

  Dataset consisting entirely of ones (in N dimensions)


Mathematical models
-------------------

Besides the primitive models listed above, there is a growing number of
mathematical models implementing comparably simple mathematical equations
that are often used. Packages derived from the ASpecD framework may well
define more specific models as well.


* :class:`aspecd.model.Polynomial`

  Polynomial (of arbitrary degree/order, depending on the number of
  coefficients)

* :class:`aspecd.model.Gaussian`

  Generalised Gaussian where amplitude, position, and width can be
  set explicitly. Hence, this is usually *not* identical to the probability
  density function (PDF) of a normally distributed random variable.

* :class:`aspecd.model.NormalisedGaussian`

  Normalised Gaussian with an integral of one, identical to the probability
  density function (PDF) of a normally distributed random variable.

* :class:`aspecd.model.Lorentzian`

  Generalised Lorentzian where amplitude, position, and width can be
  set explicitly. Hence, this is usually *not* identical to the probability
  density function (PDF) of the Cauchy distribution.

* :class:`aspecd.model.NormalisedLorentzian`

  Normalised Lorentzian with an integral of one, identical to the probability
  density function (PDF) of the Cauchy distribution.

* :class:`aspecd.model.Voigtian`

  Voigt profile: convolution of (normalised) Lorentzian and (normalised)
  Gaussian, often used to describe spectroscopic data.

* :class:`aspecd.model.Sine`

  Sine wave with adjustable amplitude, frequency, and phase.

* :class:`aspecd.model.Exponential`

  Exponential function with adjustable prefactor and rate.


Composite models consisting of a sum of individual models
---------------------------------------------------------

Often you encounter situations where a model consists of a (weighted) sum of
individual models. A simple example would be a damped oscillation. Or think
of a spectral line consisting of several overlapping individual lines
(Lorentzian or Gaussian).

All this can be easily set up using the :class:`aspecd.model.CompositeModel`
class that lets you conveniently specify a list of models, their individual
parameters, and optional weights.


Family of curves
----------------

Systematically varying one parameter at a time for a given model is key
to understanding the impact this parameter has. Therefore, automatically
creating a family of curves with one parameter varied is quite convenient.

To achieve this, use the class :class:`aspecd.model.FamilyOfCurves` that will
take the name of a model (needs to be the name of an existing model class)
and create a family of curves for this model, adding the name of the
parameter as quantity to the additional axis.


Writing your own models
=======================

All models should inherit from the :class:`aspecd.model.Model` class.
Furthermore, they should conform to a series of requirements:

* Parameters are stored in the :attr:`aspecd.model.Model.parameters` dict.

  Note that this is a :class:`dict`. In the simplest case, you may name the
  corresponding key "coefficients", as in case of a polynomial. In other
  cases, there are common names for parameters, such as "mu" and "sigma" for
  a Gaussian. Whether the keys should be named this way or describe the
  actual meaning of the parameter is partly a matter of personal taste. Use
  whatever is more common in the given context, but tend to be descriptive.
  Usually, implementing mathematical equations by simply naming every
  variable according to the mathematical notation is a bad idea, as the
  programmer will not know what these variables represent.

* Models create calculated datasets of class
  :class:`aspecd.dataset.CalculatedDataset`.

  The data of these datasets need to have dimensions corresponding to the
  variables set for the model. Think of the variables as being the axes
  values of the resulting dataset.

  The ``_origdata`` property of the dataset is automatically set accordingly
  (see below for details). This is crucially important to have the resulting
  dataset work as expected, including undo and redo functionality within the
  ASpecD framework. Remember: A calculated dataset is a regular dataset,
  and you can perform all the tasks with you would do with other datasets,
  including processing, analysis and alike.

* Model creation takes place entirely in the non-public ``_perform_task``
  method of the model.

  This method gets called from :meth:`aspecd.model.Model.create`, but not
  before some background checks have been performed, including preparing the
  metadata of the :obj:`aspecd.dataset.CalculatedDataset` object returned by
  :meth:`aspecd.model.Model.create`.

  After calling out to ``_perform_task``, the axes of the
  :obj:`aspecd.dataset.CalculatedDataset` object returned by
  :meth:`aspecd.model.Model.create` are set accordingly, *i.e.* fitting to
  the shape of the data.


On the other hand, a series of things will be automatically taken care of
for you:

* Metadata of the resulting :obj:`aspecd.dataset.CalculatedDataset` object
  are automatically set, including ``type`` (set to the full class name of
  the model) and ``parameters`` (copied over from the parameters attribute
  of the model).

* Axes of the resulting :obj:`aspecd.dataset.CalculatedDataset` object are
  automatically adjusted according to the size and content of
  the :attr:`aspecd.model.Model.variables` attribute.

  In case you used :meth:`aspecd.model.Model.from_dataset`, the axes from
  the dataset will be copied over from there.

* The ``_origdata`` property of the dataset is automatically set accordingly.
  This is crucially important to have the resulting dataset work as
  expected, including undo and redo functionality within the ASpecD framework.


Make sure your models do not raise errors such as :class:`ZeroDivisionError`
depending on the parameters set. Use the :func:`aspecd.utils.not_zero`
function where appropriate. This is particularly important in light of using
models in the context of automated fitting.


Module documentation
====================

"""

import copy

import numpy as np
import scipy.special

import aspecd.dataset
import aspecd.exceptions
import aspecd.utils
from aspecd.utils import not_zero, isiterable, ToDictMixin


class Model(ToDictMixin):
    """
    Base class for numerical models.

    Models are defined by (constant) parameters and variables the model is
    evaluated for. The variables can be thought of as the axes values of the
    resulting (calculated) dataset.

    As a simple example, consider a polynomial defined by its (constant)
    coefficients. The model will evaluate the polynomial for the values,
    and the result will be a :obj:`aspecd.dataset.CalculatedDataset` object
    containing the values of the evaluated model in its data, and the
    variables as its axes values.

    Models can be seen as abstraction to simulations in some regard. In this
    respect, they will play a central role in conjunction with fitting
    models to data by adjusting their respective parameters, a quite general
    approach in science and particularly in spectroscopy.

    Attributes
    ----------
    name : :class:`str`
        Name of the model.

        Defaults to the lower-case class name, don't change!
    parameters : :class:`dict`
        constant parameters characterising the model

    variables : :class:`list`
       values to evaluate the model for

       Usually :class:`numpy.ndarray` arrays, one for each variable

       The variables will become the values of the respective axes.

    description : :class:`str`
        Short description, to be set in class definition

    references : :class:`list`
        List of references with relevance for the implementation of the model.

        Use appropriate record types from the `bibrecord package
        <https://bibrecord.docs.till-biskup.de/>`_.

    label : :class:`str`
        Label that will be applied to the calculated dataset

        Usually, labels provide a short and concise description of a
        dataset, at least in a given context.

    axes : :class:`list`
        List of dicts containing quantity and unit for each axis.

        Needs to have the same length as the axes of the created dataset.

        If you would like to skip one axis, set it to an empty dict, None,
        or False (*i.e.*, anything that evaluates to False in Python). This
        is particularly helpful with models such as :class:`FamilyOfCurves`
        that auto-generate one axis.


    .. versionchanged:: 0.3
        New attribute :attr:`description`

    .. versionchanged:: 0.3
        New non-public method :meth:`_sanitise_parameters`

    .. versionchanged:: 0.4
        New attribute :attr:`references`

    .. versionchanged:: 0.6
        New attributes :attr:`label` and :attr:`axes`

    """

    def __init__(self):
        super().__init__()
        self.name = aspecd.utils.full_class_name(self)
        self.parameters = {}
        self.variables = []
        self.description = "Abstract model"
        self.references = []
        self.label = ""
        self.axes = []
        self._dataset = aspecd.dataset.CalculatedDataset()
        self._axes_from_dataset = []
        self.__kind__ = "model"
        self._exclude_from_to_dict = ["name", "description", "references"]

    def create(self):
        """
        Create dataset containing the evaluated model as data

        The actual model creation should be implemented within the non-public
        method :meth:`_perform_task`. Furthermore, you should make sure your
        model will be evaluated for the values given in
        :attr:`aspecd.model.Model.values` and the resulting dataset having set
        the axes appropriately.

        Furthermore, don't forget to set the ``_origdata`` property of the
        dataset, usually simply by copying the ``data`` property over there
        after it has been filled with content. This is crucially important
        to have the resulting dataset work as expected, including undo and
        redo functionality within the ASpecD framework. Remember: A
        calculated dataset is a regular dataset, and you can perform all the
        tasks with you would do with other datasets, including processing,
        analysis and alike.

        Returns
        -------
        dataset : :class:`aspecd.dataset.CalculatedDataset`
            Calculated dataset containing the evaluated model as data

        Raises
        ------
        aspecd.exceptions.MissingParameterError
            Raised if either parameters or values are not set

        """
        self._sanitise_parameters()
        self._check_prerequisites()
        self._set_dataset_metadata()
        self._perform_task()
        self._set_dataset_axes()
        self._set_dataset_origdata()
        return self._dataset

    def evaluate(self):
        """
        Evaluate model and return numerical data without any checks.

        .. important::

            Usually, you should always use :meth:`create` and obtain a
            dataset based on the model. However, :meth:`create` performs a
            lot of additional checks. Therefore, if you are sure to have set
            all properties as necessary and are interested in a probably much
            faster evaluation of the model for a given set of parameters,
            *e.g.* in context of fitting, this is the method of choice.

        Returns
        -------
        data : :class:`np.array`
            Numerical data of the model


        .. versionadded:: 0.7

        """
        self._perform_task()
        return self._dataset.data.data

    def from_dataset(self, dataset=None):
        """
        Obtain crucial information from an existing dataset.

        Often, models should be calculated for the same values as an
        existing dataset. Therefore, you can set the
        :attr:`aspecd.model.Model.values` property from a given dataset.

        If you get the variables from an existing dataset, the calculated
        dataset containing the evaluated model will have the same axes
        settings. Thus, it is pretty convenient to get a model with
        identical axes, including quantity etcetera. This helps a lot with
        plotting both, an (experimental) dataset and the model, in one plot.

        Parameters
        ----------
        dataset : :class:`aspecd.dataset.Dataset`
            Dataset to obtain crucial information for building the model from

        Raises
        ------
        aspecd.exceptions.MissingDatasetError
            Raised if no dataset is provided

        """
        if not dataset:
            raise aspecd.exceptions.MissingDatasetError
        for index in range(len(dataset.data.axes) - 1):
            self.variables.append(dataset.data.axes[index].values)
        self._axes_from_dataset = dataset.data.axes

    def from_dict(self, dict_=None):
        """
        Set attributes from dictionary.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary containing information of a task.

        Raises
        ------
        aspecd.plotting.MissingDictError
            Raised if no dict is provided.

        """
        if not dict_:
            raise aspecd.exceptions.MissingDictError(
                "Need a dict to read from, but none given"
            )
        for key, value in dict_.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def _check_prerequisites(self):
        if not self.parameters:
            raise aspecd.exceptions.MissingParameterError(
                "No parameters for model provided"
            )
        if len(self.variables) == 0:
            raise aspecd.exceptions.MissingParameterError(
                "No variables to evaluate model for provided"
            )

    def _sanitise_parameters(self):
        """Ensure parameters provided for model are correct.

        Needs to be implemented in classes inheriting from Model
        according to their needs. Most probably, you want to check for
        correct types of all parameters as well as values within sensible
        borders.

        """

    def _perform_task(self):
        """Create the actual model and evaluate it for the given values.

        The implementation of the actual model goes in here in all
        classes inheriting from Model. This method is automatically
        called by :meth:`self.create` after some background checks.

        """
        # dummy to get tests to run
        self._dataset.data.data = self.variables[0]

    def _set_dataset_axes(self):
        """
        Set axes of calculated dataset

        In case a dataset has been used to obtain the variables for,
        the axes get set from this dataset.
        """
        if self._axes_from_dataset:
            self._dataset.data.axes = self._axes_from_dataset
        else:
            if isinstance(self.variables[0], (list, np.ndarray)):
                for index, _ in enumerate(self.variables):
                    self._dataset.data.axes[index].values = self.variables[
                        index
                    ]
            else:
                self._dataset.data.axes[0].values = np.asarray(self.variables)
            self._dataset.data.axes[-1].quantity = "amplitude"
            self._dataset.data.axes[-1].unit = "a.u."
        if self.axes:
            if len(self.axes) != len(self._dataset.data.axes):
                message = (
                    "Number of axes and dataset axes need to be identical"
                )
                raise IndexError(message)
            for idx, axis in enumerate(self._dataset.data.axes):
                if self.axes[idx]:
                    axis.quantity = self.axes[idx]["quantity"]
                    axis.unit = self.axes[idx]["unit"]

    def _set_dataset_metadata(self):
        """
        Set calculation metadata of calculated dataset

        Calculation type is set to the full class name of the respective
        model, and parameters are copied over from the model parameters.
        """
        self._dataset.metadata.calculation.type = self.name
        self._dataset.metadata.calculation.parameters = self.parameters
        self._dataset.label = self.label

    def _set_dataset_origdata(self):
        # pylint: disable=protected-access
        self._dataset._origdata = copy.deepcopy(self._dataset.data)


class CompositeModel(Model):
    """
    Composite model consisting of a weighted contributions of individual models.

    Individual models can either be added up (default) or multiplied,
    depending on which operators are provided. Both situations occur
    frequently. If you would like to describe a spectrum as sum of Gaussian
    or Lorentzian lines, you need to add the individual contributions. If
    you would like to model a damped oscillation, you would need to multiply
    the exponential decay onto the oscillation.

    Attributes
    ----------
    models : :class:`list`
        Names of the models the composite model consists of

        Each name needs to be the name of an existing model class.

    parameters : :class:`list`
        Constant parameters characterising each individual model

        For the parameters that can (and need to) be set, consult the
        documentation of each of the respective model classes specified in
        the :attr:`models` attribute.

    weights : :class:`list`
        Factors used to weight the individual models.

        Default: no weighting

    operators : :class:`list`
        Operators to be used for the individual models.

        Addition ("+", "add", "plus") and multiplication ("*", "multiply",
        "times") are supported.

        Note that one operator less than models needs to be provided.

        Default: add


    Raises
    ------
    IndexError
        Raised if number of models, parameter sets, operators, and weights are
        incompatible


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Suppose you would want to describe your data with a model consisting of
    two Lorentzian line shapes. Starting from scratch, you need to create a
    dummy dataset (using, *e.g.*, :class:`aspecd.model.Zeros`) of given
    length and axes range. Based on that you can create your model:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1001
             range: [0, 20]
         result: dummy

       - kind: model
         type: CompositeModel
         from_dataset: dummy
         properties:
           models:
             - Lorentzian
             - Lorentzian
           parameters:
             - position: 5
             - position: 8
         result: multiple_lorentzians

    Note that you need to provide parameters for each of the individual
    models, even if the class for a model would work without explicitly
    providing parameters.

    Of course, if you start with an existing dataset (*e.g.*, loaded from
    some real data), you could use the label to this dataset directly in
    ``from_dataset``, without needing to create a dummy dataset first.

    While adding up the contributions of the individual components works
    well for describing spectra, sometimes you need to multiply contributions.
    Suppose you would want to create a damped oscillation consisting of a
    sine and an exponential. Starting from scratch, you need to create a
    dummy dataset (using, *e.g.*, :class:`aspecd.model.Zeros`) of given
    length and axes range. Based on that you can create your model:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1001
             range: [0, 20]
         result: dummy

       - kind: model
         type: CompositeModel
         from_dataset: dummy
         properties:
           models:
             - Sine
             - Exponential
           parameters:
             - frequency: 1
               phase: 1.57
             - rate: -0.2
           operators:
             - multiply
         result: damped_oscillation

    Again, you need to provide parameters for each of the individual
    models, even if the class for a model would work without explicitly
    providing parameters.


    .. versionadded:: 0.3

    """

    def __init__(self):
        super().__init__()
        self.description = (
            "Composite model consisting of several weighted models"
        )
        self.models = []
        self.parameters = []
        self.weights = []
        self.operators = []

    def _sanitise_parameters(self):
        if not self.weights:
            self.weights = np.ones(len(self.models))
        if not self.operators:
            for _ in self.models:
                self.operators.append("+")
        else:
            self.operators.insert(0, "+")
        if len(self.parameters) != len(self.models):
            raise IndexError("Models and parameters count differs")
        if len(self.weights) != len(self.models):
            raise IndexError("Models and weights count differs")
        if len(self.operators) != len(self.models):
            raise IndexError("Models and operators count differs")

    def _perform_task(self):
        if isinstance(self.variables, list):
            data = np.zeros([len(x) for x in self.variables])
        else:
            data = np.zeros(len(self.variables))
        for idx, model_name in enumerate(self.models):
            model = self._get_model(model_name)
            for key in self.parameters[idx]:
                # noinspection PyUnresolvedReferences
                model.parameters[key] = self.parameters[idx][key]
            model.variables = self.variables
            # noinspection PyUnresolvedReferences
            dataset = model.create()
            data = data.reshape(dataset.data.data.shape)
            if self.operators[idx] in ("+", "plus", "add"):
                data += dataset.data.data * self.weights[idx]
            if self.operators[idx] in ("*", "times", "multiply"):
                data *= dataset.data.data * self.weights[idx]
        self._dataset.data.data = data

    @staticmethod
    def _get_model(model_name):
        try:
            model = aspecd.utils.object_from_class_name(model_name)
        except (ValueError, AttributeError):
            model = aspecd.utils.object_from_class_name(
                "aspecd.model." + model_name
            )
        return model


class FamilyOfCurves(Model):
    """
    Create a family of curves for a model, varying a single parameter.

    Systematically varying one parameter at a time for a given model is key
    to understanding the impact this parameter has. Therefore, automatically
    creating a family of curves with one parameter varied is quite convenient.

    This class will take the name of a model (needs to be the name of an
    existing model class) and create a family of curves for this model,
    adding the name of the parameter as quantity to the additional axis.


    Attributes
    ----------
    model : :class:`str`
        Name of the model the family of curves should be calculated for

        Needs to be the name of an existing model class.

    vary : :class:`dict`
        Name and values of the parameter to be varied

        parameter : :class:`str`
            Name of the parameter that should be varied

        values : :class:`list`
            Values of the parameter to be varied


    Raises
    ------
    ValueError
        Raised if no model is provided


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Suppose you would want to create a family of curves of a Gaussian with
    varying the width. Starting from scratch, you need to create a
    dummy dataset (using, *e.g.*, :class:`aspecd.model.Zeros`) of given
    length and axes range. Based on that you can create your family of curves:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1001
             range: [-5, 5]
         result: dummy

       - kind: model
         type: FamilyOfCurves
         from_dataset: dummy
         properties:
           model: Gaussian
           vary:
             parameter: width
             values: [1., 1.5, 2., 2.5, 3]
         result: gaussian_with_varied_width

    This would create a 2D dataset with a Gaussian with standard values for
    amplitude and position and the value for the width varied as given.

    Of course, if you start with an existing dataset (*e.g.*, loaded from
    some real data), you could use the label to this dataset directly in
    ``from_dataset``, without needing to create a dummy dataset first.

    If you would like to control additional parameters of the Gaussian,
    you can do that as well:

    .. code-block:: yaml

       - kind: model
         type: FamilyOfCurves
         from_dataset: dummy
         properties:
           model: Gaussian
           parameters:
             amplitude: 3.
             position: -1
           vary:
             parameter: width
             values: [1., 1.5, 2., 2.5, 3]
         result: gaussian_with_varied_width

    Note that if you provide a value for the parameter to be varied in the
    list of parameters, it will be silently overwritten by the values
    provided with ``vary``.


    .. versionadded:: 0.3

    """

    def __init__(self):
        super().__init__()
        self.description = (
            "Family of curves for a model with one parameter varied"
        )
        self.model = None
        self.vary = {}

    def _sanitise_parameters(self):
        if not self.model:
            raise ValueError("Missing a model")
        if not isiterable(self.vary["values"]):
            self.vary["values"] = [self.vary["values"]]
        if not self.parameters:
            self.parameters[self.vary["parameter"]] = self.vary["values"][0]

    # noinspection PyUnresolvedReferences
    def _perform_task(self):
        self._dataset.data.data = np.zeros(
            [len(self.variables[0]), len(self.vary["values"])]
        )
        model = self._get_model(self.model)
        model.variables = self.variables
        for key, value in self.parameters.items():
            model.parameters[key] = value
        for idx, value in enumerate(self.vary["values"]):
            model.parameters[self.vary["parameter"]] = value
            dataset = model.create()
            self._dataset.data.data[:, idx] = dataset.data.data
        if len(self.vary["values"]) > 1:
            self._dataset.data.axes[-2].quantity = self.vary["parameter"]
            self._dataset.data.axes[-2].values = [
                float(x) for x in self.vary["values"]
            ]
        if self._axes_from_dataset:
            self._axes_from_dataset.insert(-1, self._dataset.data.axes[-2])

    @staticmethod
    def _get_model(model_name):
        try:
            model = aspecd.utils.object_from_class_name(model_name)
        except (ValueError, AttributeError):
            model = aspecd.utils.object_from_class_name(
                "aspecd.model." + model_name
            )
        return model


class Zeros(Model):
    # noinspection PyUnresolvedReferences
    """
    Zeros of given shape.

    One of the most primitive models: zeros in N dimensions.

    This model is quite helpful for creating test datasets, *e.g.* with
    added noise (of different colour). Basically, it can be thought of as a
    wrapper for :func:`numpy.zeros`. Its particular strength is that using
    this model, creating test datasets becomes straight-forward in context
    of recipe-driven data analysis.


    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        shape : :class:`list`
            shape of the data

            Have in mind that ND datasets get huge very fast. Therefore,
            it is *not* the best idea to create an 3D dataset with zeros
            with 2**12 elements along each dimension.

        range : :class:`list`
            range of each of the axes

            Useful if you want to specify the axes values as well.

            If the data are multidimensional, one range for each axis needs
            to be provided.

    Raises
    ------
    aspecd.exceptions.MissingParameterError
        Raised if no shape is given

    IndexError
        Raised if elements in shape and range are incompatible


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Creating a dataset consisting of 2**10 zeros is quite simple:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1024
         result: 1d_zeros

    Of course, you are not limited to 1D datasets, and you can easily create
    ND datasets as well:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: [1024, 256, 256]
         result: 3d_zeros

    Please have in mind that the memory of your computer is usually limited
    and that ND datasets become huge very fast. Hence, creating a 3D array
    with 2**10 elements along each dimension is most probably *not* the best
    idea.

    Suppose you not only want to create a dataset with a given shape,
    but set the axes values (*i.e.*, their range) as well:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1024
             range: [35, 42]
         result: 1d_zeros

    This would create a 1D dataset with 1024 values, with the axes
    values spanning a range from 35 to 42. Of course, the same can be done
    with ND datasets.

    Now, let's assume that you would want to play around with the different
    types of (coloured) noise. Therefore, you would want to first create a
    dataset and afterwards add noise to it:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 8192
         result: 1d_zeros

       - kind: processing
         type: Noise
         properties:
           parameters:
             normalise: True

    This would create a dataset consisting of 2**14 zeros and add pink
    (1/*f*) noise to it that is normalised (has an amplitude of 1). To check
    that the noise is really 1/*f* noise, you may look at its power density.
    See :class:`aspecd.analysis.PowerDensitySpectrum` for details, including
    how to even plot both, the power density spectrum and a linear fit
    together in one figure.


    .. versionadded:: 0.3

    """

    def __init__(self):
        super().__init__()
        self.description = "Model containing only zeros"
        self.parameters["shape"] = None
        self.parameters["range"] = None

    def _sanitise_parameters(self):
        if not self.variables:
            self.variables = [0]
            if not self.parameters["shape"]:
                raise aspecd.exceptions.MissingParameterError(
                    message="Parameter 'shape' missing"
                )
        if not self.parameters["shape"]:
            self.parameters["shape"] = []
            if isiterable(self.variables[0]):
                for index, _ in enumerate(self.variables):
                    # noinspection PyTypeChecker
                    self.parameters["shape"].append(
                        len(self.variables[index])
                    )
            else:
                self.parameters["shape"] = len(self.variables)
        if not isiterable(self.parameters["shape"]):
            self.parameters["shape"] = [self.parameters["shape"]]
        if self.parameters["range"]:
            if not isiterable(self.parameters["range"][0]):
                self.parameters["range"] = [self.parameters["range"]]
            if len(self.parameters["shape"]) != len(self.parameters["range"]):
                raise IndexError("Shape and range must be compatible")

    def _perform_task(self):
        self._dataset.data.data = np.zeros(self.parameters["shape"])
        if self.parameters["range"]:
            self._set_variables()

    def _set_variables(self):
        self.variables = []
        shape = self.parameters["shape"]
        range_ = self.parameters["range"]
        for dim in range(self._dataset.data.data.ndim):
            axis_values = np.linspace(
                range_[dim][0], range_[dim][1], shape[dim]
            )
            self.variables.append(axis_values)


class Ones(Model):
    # noinspection PyUnresolvedReferences
    """
    Ones of given shape.

    One of the most primitive models: ones in N dimensions.

    This model is quite helpful for creating test datasets, *e.g.* with
    added noise (of different colour). Basically, it can be thought of as a
    wrapper for :func:`numpy.ones`. Its particular strength is that using
    this model, creating test datasets becomes straight-forward in context
    of recipe-driven data analysis.


    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        shape : :class:`list`
            shape of the data

            Have in mind that ND datasets get huge very fast. Therefore,
            it is *not* the best idea to create an 3D dataset with ones
            with 2**12 elements along each dimension.

        range : :class:`list`
            range of each of the axes

            Useful if you want to specify the axes values as well.

            If the data are multidimensional, one range for each axis needs
            to be provided.

    Raises
    ------
    aspecd.exceptions.MissingParameterError
        Raised if no shape is given

    IndexError
        Raised if elements in shape and range are incompatible


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Creating a dataset consisting of 2**10 ones is quite simple:

    .. code-block:: yaml

       - kind: model
         type: Ones
         properties:
           parameters:
             shape: 1024
         result: 1d_ones

    Of course, you are not limited to 1D datasets, and you can easily create
    ND datasets as well:

    .. code-block:: yaml

       - kind: model
         type: Ones
         properties:
           parameters:
             shape: [1024, 256, 256]
         result: 3d_ones

    Please have in mind that the memory of your computer is usually limited
    and that ND datasets become huge very fast. Hence, creating a 3D array
    with 2**10 elements along each dimension is most probably *not* the best
    idea.

    Suppose you not only want to create a dataset with a given shape,
    but set the axes values (*i.e.*, their range) as well:

    .. code-block:: yaml

       - kind: model
         type: Ones
         properties:
           parameters:
             shape: 1024
             range: [35, 42]
         result: 1d_zeros

    This would create a 1D dataset with 1024 values, with the axes
    values spanning a range from 35 to 42. Of course, the same can be done
    with ND datasets.

    Now, let's assume that you would want to play around with the different
    types of (coloured) noise. Therefore, you would want to first create a
    dataset and afterwards add noise to it:

    .. code-block:: yaml

       - kind: model
         type: Ones
         properties:
           parameters:
             shape: 8192
         result: 1d_ones

       - kind: processing
         type: Noise
         properties:
           parameters:
             normalise: True

    This would create a dataset consisting of 2**14 ones and add pink
    (1/*f*) noise to it that is normalised (has an amplitude of 1). To check
    that the noise is really 1/*f* noise, you may look at its power density.
    See :class:`aspecd.analysis.PowerDensitySpectrum` for details, including
    how to even plot both, the power density spectrum and a linear fit
    together in one figure.


    .. versionadded:: 0.3

    """

    def __init__(self):
        super().__init__()
        self.description = "Model containing only ones"
        self.parameters["shape"] = None
        self.parameters["range"] = None

    def _sanitise_parameters(self):
        if not self.variables:
            self.variables = [0]
            if not self.parameters["shape"]:
                raise aspecd.exceptions.MissingParameterError(
                    message="Parameter 'shape' missing"
                )
        if not self.parameters["shape"]:
            self.parameters["shape"] = []
            if isiterable(self.variables[0]):
                for index, _ in enumerate(self.variables):
                    # noinspection PyTypeChecker
                    self.parameters["shape"].append(
                        len(self.variables[index])
                    )
            else:
                self.parameters["shape"] = len(self.variables)
        if not isiterable(self.parameters["shape"]):
            self.parameters["shape"] = [self.parameters["shape"]]
        if self.parameters["range"]:
            if not isiterable(self.parameters["range"][0]):
                self.parameters["range"] = [self.parameters["range"]]
            if len(self.parameters["shape"]) != len(self.parameters["range"]):
                raise IndexError("Shape and range must be compatible")

    def _perform_task(self):
        self._dataset.data.data = np.ones(self.parameters["shape"])
        if self.parameters["range"]:
            self._set_variables()

    def _set_variables(self):
        self.variables = []
        shape = self.parameters["shape"]
        range_ = self.parameters["range"]
        for dim in range(self._dataset.data.data.ndim):
            axis_values = np.linspace(
                range_[dim][0], range_[dim][1], shape[dim]
            )
            self.variables.append(axis_values)


class Polynomial(Model):
    # noinspection PyUnresolvedReferences
    """
    Polynomial.

    Evaluate a polynomial with given coefficients for the data provided in
    :attr:`aspecd.model.Model.variables`.

    .. note::
        As the new :mod:`numpy.polynomial` package is used, particularly
        the :class:`numpy.polynomial.polynomial.Polynomial` class,
        the coefficients are given in increasing order, with the first
        element corresponding to x**0.

        Furthermore, the coefficients are assumed to be provided in the
        unscaled data domain (by using the
        :meth:`numpy.polynomial.polynomial.Polynomial.convert` method).


    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        coefficients : :class:`list`
            coefficients of the polynomial to be evaluated

            The number of coefficients determines the order (degree) of the
            polynomial. The coefficients have to be given in *increasing*
            order (see note above). Furthermore, you need to provide the
            coefficients in the unscaled data domain (using the
            :meth:`numpy.polynomial.polynomial.Polynomial.convert` method).

    Raises
    ------
    aspecd.exceptions.MissingParameterError
        Raised if no coefficients are given


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Suppose you would want to create a Polynomial of first order with a
    slope of 42 and an intercept of -3. Starting from scratch, you need to
    create a dummy dataset (using, *e.g.*, :class:`aspecd.model.Zeros`) of
    given length and axes range. Based on that you can create your Polynomial:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1001
             range: [-5, 5]
         result: dummy

       - kind: model
         type: Polynomial
         from_dataset: dummy
         properties:
           parameters:
             coefficients: [-3, 42]
         result: polynomial

    Note that the coefficients are given in *increasing* order of the
    exponent, here intercept first, followed by the slope.

    Of course, if you start with an existing dataset (*e.g.*, loaded from
    some real data), you could use the label to this dataset directly in
    ``from_dataset``, without needing to create a dummy dataset first.


    .. versionadded:: 0.3

    """

    def __init__(self):
        super().__init__()
        self.description = "Polynomial"
        self.parameters["coefficients"] = None

    def _sanitise_parameters(self):
        if not self.parameters["coefficients"]:
            raise aspecd.exceptions.MissingParameterError(
                message="Parameter 'coefficients' missing"
            )

    def _perform_task(self):
        polynomial = np.polynomial.Polynomial(self.parameters["coefficients"])
        self._dataset.data.data = polynomial(self.variables[0])


class Gaussian(Model):
    # noinspection PyUnresolvedReferences
    r"""
    Generalised Gaussian.

    Creates a Gaussian function or Gaussian, with its characteristic
    symmetric "bell curve" shape.

    The underlying mathematical equation may be written as follows:

    .. math::

        f(x) = a \exp\left(-\frac{(x-b)^2}{2c^2}\right)

    with :math:`a` being the amplitude, :math:`b` the position,
    and :math:`c` the width of the Gaussian.

    .. important::
        Note that this is a generalised Gaussian where you can set
        amplitude, position, and width independently. Hence, it is *not*
        normalised to an integral of one, and therefore *not* to be confused
        with the the probability density function (PDF) of a normally
        distributed random variable. If you are interested in this, see the
        :class:`aspecd.model.NormalisedGaussian` class.


    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        amplitude : :class:`float`
            Amplitude or height of the Gaussian

            Default: 1

        position : :class:`float`
            Position (of the maximum) of the Gaussian

            Default: 0

        width : :class:`float`
            Width of the Gaussian

            The full width at half maximum (FWHM) is related to the width
            *b* by: :math:`2 \sqrt{2 \log(2)} b`.

            Default: 1


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Suppose you would want to create a Gaussian with standard values
    (amplitude=1, position=0, width=1). Starting from scratch, you need to
    create a dummy dataset (using, *e.g.*, :class:`aspecd.model.Zeros`) of
    given length and axes range. Based on that you can create your Gaussian:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1001
             range: [-5, 5]
         result: dummy

       - kind: model
         type: Gaussian
         from_dataset: dummy
         result: gaussian

    Of course, if you start with an existing dataset (*e.g.*, loaded from
    some real data), you could use the label to this dataset directly in
    ``from_dataset``, without needing to create a dummy dataset first.

    Of course, you can control all three parameters (amplitude, position,
    width) explicitly:

    .. code-block:: yaml

       - kind: model
         type: Gaussian
         properties:
           parameters:
             amplitude: 5
             position: 1.5
             width: 0.5
         from_dataset: dummy
         result: gaussian

    This would create a Gaussian with an amplitude (height) of 5, situated
    at a value of 1.5 at the *x* axis, and with a width of 0.5.

    .. versionadded:: 0.3

    """

    def __init__(self):
        super().__init__()
        self.description = "Generalised Gaussian"
        self.parameters["amplitude"] = 1
        self.parameters["position"] = 0
        self.parameters["width"] = 1

    def _perform_task(self):
        x = np.asarray(self.variables[0])  # pylint: disable=invalid-name
        amplitude = self.parameters["amplitude"]
        position = self.parameters["position"]
        width = self.parameters["width"]
        gaussian = amplitude * np.exp(
            -((x - position) ** 2) / not_zero(2 * width**2)
        )
        self._dataset.data.data = gaussian


class NormalisedGaussian(Model):
    # noinspection PyUnresolvedReferences
    r"""
    Normalised Gaussian.

    Creates a Gaussian function or Gaussian, with its characteristic
    symmetric "bell curve" shape, normalised to an integral of one. Thus,
    it is the probability density function (PDF) of a normally distributed
    random variable.

    The underlying mathematical equation may be written as follows:

    .. math::

        f(x) = \frac{1}{\sigma\sqrt{2\pi}} \exp\left(-\frac{(x-\mu)^2}{
        2\sigma^2}\right)

    with :math:`\mu` being the position and :math:`\sigma` the width of the
    Gaussian, and :math:`\sigma^2` the variance.

    .. note::

        This class creates a normalised Gaussian, equivalent to the PDF of a
        normally distributed random variable. If you are interested in a
        Gaussian where you can set all three parameters (amplitude,
        position, width) independently, see the
        :class:`aspecd.model.Gaussian` class.


    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        position : :class:`float`
            Position (of the maximum) of the Gaussian

            For a normally distributed random variable :math:`x`,
            the position is identical to its expected value :math:`E(x)` or
            mean :math:`\mu`. Other names include first moment and average.

            Default: 0

        width : :class:`float`
            Width of the Gaussian

            The full width at half maximum (FWHM) is related to the width
            :math:`\sigma` by: :math:`2 \sqrt{2 \log(2)} \sigma`. The
            squared value of the width is better known the variance
            :math:`\sigma^2`.

            Default: 1


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Suppose you would want to create a normalised Gaussian with standard values
    (position=0, width=1). Starting from scratch, you need to create a dummy
    dataset (using, *e.g.*, :class:`aspecd.model.Zeros`) of given length and
    axes range. Based on that you can create your normalised Gaussian:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1001
             range: [-5, 5]
         result: dummy

       - kind: model
         type: NormalisedGaussian
         from_dataset: dummy
         result: normalised_gaussian

    Of course, if you start with an existing dataset (*e.g.*, loaded from
    some real data), you could use the label to this dataset directly in
    ``from_dataset``, without needing to create a dummy dataset first.

    Of course, you can control position and width explicitly:

    .. code-block:: yaml

       - kind: model
         type: NormalisedGaussian
         properties:
           parameters:
             position: 1.5
             width: 0.5
         from_dataset: dummy
         result: normalised_gaussian

    This would create a normalised Gaussian with its maximum situated
    at a value of 1.5 at the *x* axis, and with a width of 0.5.

    .. versionadded:: 0.3

    """

    def __init__(self):
        super().__init__()
        self.description = "Normalised Gaussian"
        self.parameters["position"] = 0
        self.parameters["width"] = 1

    def _perform_task(self):
        x = np.asarray(self.variables[0])  # pylint: disable=invalid-name
        position = self.parameters["position"]
        width = self.parameters["width"]
        amplitude = 1 / not_zero(width * np.sqrt(2 * np.pi))
        gaussian = amplitude * np.exp(
            -((x - position) ** 2) / not_zero(2 * width**2)
        )
        self._dataset.data.data = gaussian


class Lorentzian(Model):
    # noinspection PyUnresolvedReferences
    r"""
    Generalised Lorentzian.

    Creates a Lorentzian function or Lorentzian often used in spectroscopy,
    as the line shape of a purely lifetime-broadened spectral line is
    identical to such a Lorentzian.

    The underlying mathematical equation may be written as follows:

    .. math::

        f(x) = a \left[\frac{c^2}{(x-b)^2 + c^2}\right]

    with :math:`a` being the amplitude, :math:`b` the position,
    and :math:`c` the width of the Lorentzian.

    .. important::
        Note that this is a generalised Lorentzian where you can set
        amplitude, position, and width independently. Hence, it is *not*
        normalised to an integral of one, and therefore *not* to be confused
        with the the probability density function (PDF) of the Cauchy
        distribution. If you are interested in this, see the
        :class:`aspecd.model.NormalisedLorentzian` class.


    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        amplitude : :class:`float`
            Amplitude or height of the Lorentzian

            Default: 1

        position : :class:`float`
            Position (of the maximum) of the Lorentzian

            Default: 0

        width : :class:`float`
            Width of the Lorentzian

            The full width at half maximum (FWHM) is related to the width
            :math:`b` by: :math:`2b`.

            Default: 1


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Suppose you would want to create a Lorentzian with standard values
    (amplitude=1, position=0, width=1). Starting from scratch, you need to
    create a dummy dataset (using, *e.g.*, :class:`aspecd.model.Zeros`) of
    given length and axes range. Based on that you can create your Lorentzian:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1001
             range: [-5, 5]
         result: dummy

       - kind: model
         type: Lorentzian
         from_dataset: dummy
         result: lorentzian

    Of course, if you start with an existing dataset (*e.g.*, loaded from
    some real data), you could use the label to this dataset directly in
    ``from_dataset``, without needing to create a dummy dataset first.

    Of course, you can control all three parameters (amplitude, position,
    width) explicitly:

    .. code-block:: yaml

       - kind: model
         type: Lorentzian
         properties:
           parameters:
             amplitude: 5
             position: 1.5
             width: 0.5
         from_dataset: dummy
         result: lorentzian

    This would create a Lorentzian with an amplitude (height) of 5, situated
    at a value of 1.5 at the *x* axis, and with a width of 0.5.

    .. versionadded:: 0.3

    """

    def __init__(self):
        super().__init__()
        self.description = "Generalised Lorentzian"
        self.parameters["amplitude"] = 1
        self.parameters["position"] = 0
        self.parameters["width"] = 1

    def _perform_task(self):
        x = np.asarray(self.variables[0])  # pylint: disable=invalid-name
        amplitude = self.parameters["amplitude"]
        position = self.parameters["position"]
        width = self.parameters["width"]
        lorentzian = amplitude * (
            width**2 / ((x - position) ** 2 + not_zero(width**2))
        )
        self._dataset.data.data = lorentzian


class NormalisedLorentzian(Model):
    # noinspection PyUnresolvedReferences
    r"""
    Normalised Lorentzian.

    Creates a normalised Lorentzian function or Lorentzian with an integral
    of one, *i.e.* the probability density function (PDF) of the Cauchy
    distribution.

    The underlying mathematical equation may be written as follows:

    .. math::

        f(x) = \frac{1}{\pi c} \left[\frac{c^2}{(x-b)^2 + c^2}\right] =
        \frac{c}{\pi[(x-b)^2 + c^2]}

    with :math:`b` being the position and :math:`c` the width of the
    Lorentzian.

    .. note::

        This class creates a normalised Lorentzian, equivalent to the PDF of
        the Cauchy distribution. If you are interested in a Lorentzian where
        you can set all three parameters (amplitude, position, width)
        independently, see the :class:`aspecd.model.Lorentzian` class.


    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        position : :class:`float`
            Position (of the maximum) of the Lorentzian

            Default: 0

        width : :class:`float`
            Width of the Lorentzian

            The full width at half maximum (FWHM) is related to the width
            :math:`b` by: :math:`2b`.

            Default: 1


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Suppose you would want to create a normalised Lorentzian with standard
    values (position=0, width=1). Starting from scratch, you need to create
    a dummy dataset (using, *e.g.*, :class:`aspecd.model.Zeros`) of given
    length and axes range. Based on that you can create your Lorentzian:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1001
             range: [-5, 5]
         result: dummy

       - kind: model
         type: NormalisedLorentzian
         from_dataset: dummy
         result: normalised_lorentzian

    Of course, if you start with an existing dataset (*e.g.*, loaded from
    some real data), you could use the label to this dataset directly in
    ``from_dataset``, without needing to create a dummy dataset first.

    Of course, you can control position and width explicitly:

    .. code-block:: yaml

       - kind: model
         type: NormalisedLorentzian
         properties:
           parameters:
             position: 1.5
             width: 0.5
         from_dataset: dummy
         result: normalised_lorentzian

    This would create a normalised Lorentzian with its maximum situated at a
    value of 1.5 at the *x* axis, and with a width of 0.5.

    .. versionadded:: 0.3

    """

    def __init__(self):
        super().__init__()
        self.description = (
            "Normalised Lorentzian, PDF of a Cauchy distribution"
        )
        self.parameters["position"] = 0
        self.parameters["width"] = 1

    def _perform_task(self):
        x = np.asarray(self.variables[0])  # pylint: disable=invalid-name
        position = self.parameters["position"]
        width = self.parameters["width"]
        amplitude = 1 / (np.pi * not_zero(width))
        lorentzian = amplitude * (
            width**2 / ((x - position) ** 2 + not_zero(width**2))
        )
        self._dataset.data.data = lorentzian


class Voigtian(Model):
    """
    Voigt profile.

    The Voigt profile (after Woldemar Voigt) is a probability distribution
    given by a convolution of a Cauchy-Lorentz distribution (with half-width
    at half-maximum ``gamma``) and a Gaussian distribution (with standard
    deviation ``sigma``). It is often used for analyzing spectroscopic data.

    In spectroscopy, a Voigt profile results from the convolution of two
    broadening mechanisms: life-time broadening (Lorentzian part) and
    inhomogeneous broadening (Gaussian part).

    If ``sigma = 0``, the PDF of the Cauchy distribution is returned.
    Conversely, if ``gamma = 0``, the PDF of the Normal distribution is
    returned. If ``sigma = gamma = 0`, the return value is ``Inf`` for
    ``x = 0``, and ``0`` for all other ``x``.

    Note: Internally, the function :func:`scipy.special.voigt_profile` is
    used to calculate the data.

    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        position : :class:`float`
            Position (of the maximum) of the Voigt profile

            Default: 0

        sigma : :class:`float`
            Standard deviation of the Gaussian part

            Default: 1

        gamma : :class:`float`
            Width of the Lorentzian part

            The full width at half maximum (FWHM) is related to the width
            :math:`b` by: :math:`2b`.

            Default: 1


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Suppose you would want to create a Voigt profile with standard
    values (position=0, gamma=1, sigma=1). Starting from scratch, you need to
    create a dummy dataset (using, *e.g.*, :class:`aspecd.model.Zeros`) of
    given length and axes range. Based on that you can create your Voigtian:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1001
             range: [-5, 5]
         result: dummy

       - kind: model
         type: Voigtian
         from_dataset: dummy
         result: voigtian

    Of course, if you start with an existing dataset (*e.g.*, loaded from
    some real data), you could use the label to this dataset directly in
    ``from_dataset``, without needing to create a dummy dataset first.

    Of course, you can control position and widths of Gaussian and
    Lorentzian contributions explicitly:

    .. code-block:: yaml

       - kind: model
         type: Voigtian
         properties:
           parameters:
             position: 1.5
             sigma: 0.5
             gamma: 2
         from_dataset: dummy
         result: voigtian

    This would create a Voigt profile with its maximum situated at a
    value of 1.5 at the *x* axis, and with a standard deviation of
    the Gaussian component of 0.5 and a line width of the Lorentzian part of 2.

    .. versionadded:: 0.10


    """

    def __init__(self):
        super().__init__()
        self.description = "Voigt profile"
        self.parameters["sigma"] = 1
        self.parameters["gamma"] = 1
        self.parameters["position"] = 0

    def _perform_task(self):
        x = np.asarray(self.variables[0])  # pylint: disable=invalid-name
        voigtian = scipy.special.voigt_profile(
            x - self.parameters["position"],
            self.parameters["sigma"],
            self.parameters["gamma"],
        )
        self._dataset.data.data = voigtian


class Sine(Model):
    # noinspection PyUnresolvedReferences
    r"""
    Sine wave.

    Creates a sine function with given amplitude, frequency, and phase.

    The underlying mathematical equation may be written as follows:

    .. math::

        f(x) = a \sin(fx + \phi)

    with :math:`a` being the amplitude, :math:`f` the frequency,
    and :math:`\phi` the phase of the sine.


    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        amplitude : :class:`float`
            Amplitude of the sine.

            Note that the real amplitude (max-min) is twice the value given
            here. Nevertheless, calling this factor "amplitude" seems to be
            common.

            Default: 1

        frequency : :class:`float`
            Frequency of the sine (in radians).

            Default: 1

        phase : :class:`float`
            Phase (*i.e.*, shift) of the sine (in radians).

            Setting the phase to :math:`\pi/2` would result in a cosine.

            Default: 0


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Suppose you would want to create a sine with standard values (amplitude=1,
    frequency=1, shift=0). Starting from scratch, you need to create
    a dummy dataset (using, *e.g.*, :class:`aspecd.model.Zeros`) of given
    length and axes range. Based on that you can create your sine:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1001
             range: [-5, 5]
         result: dummy

       - kind: model
         type: Sine
         from_dataset: dummy
         result: sine

    Of course, if you start with an existing dataset (*e.g.*, loaded from
    some real data), you could use the label to this dataset directly in
    ``from_dataset``, without needing to create a dummy dataset first.

    Of course, you can control all three parameters (amplitude, frequency,
    and shift) explicitly:

    .. code-block:: yaml

       - kind: model
         type: Sine
         properties:
           parameters:
             amplitude: 42
             frequency: 4.2
             shift: 1.57
         from_dataset: dummy
         result: sine

    This would create a sine with an amplitude of 42 (the actual amplitude,
    defined as max-min, would be twice this value), a frequency of 4.2 and a
    shift of about pi/2.

    .. versionadded:: 0.3

    """

    def __init__(self):
        super().__init__()
        self.description = "Sine function"
        self.parameters["amplitude"] = 1
        self.parameters["frequency"] = 1
        self.parameters["phase"] = 0

    def _perform_task(self):
        x = np.asarray(self.variables[0])  # pylint: disable=invalid-name
        amplitude = self.parameters["amplitude"]
        frequency = self.parameters["frequency"]
        phase = self.parameters["phase"]
        sine = amplitude * np.sin(frequency * x + phase)
        self._dataset.data.data = sine


class Exponential(Model):
    # noinspection PyUnresolvedReferences
    r"""
    Exponential function.

    Creates an exponential with given prefactor and rate.

    The underlying mathematical equation may be written as follows:

    .. math::

        f(x) = a \exp(bx)

    with :math:`a` being the prefactor and :math:`b` the rate of the
    exponential.


    Attributes
    ----------
    parameters : :class:`dict`
        All parameters necessary for this step.

        prefactor : :class:`float`
            Intercept of the exponential.

            Default: 1

        rate : :class:`float`
            Rate of the exponential.

            Default: 1


    .. note::
        In case of modelling exponential decays, the rate constant will
        become negative. This rate constant (decay rate) is the inverse of the
        lifetime. Lifetime and half-life are related by a factor of ln(2).


    Examples
    --------
    For convenience, a series of examples in recipe style (for details of
    the recipe-driven data analysis, see :mod:`aspecd.tasks`) is given below
    for how to make use of this class. The examples focus each on a single
    aspect.

    Suppose you would want to create an exponential with standard values
    (prefactor=1, rate=1). Starting from scratch, you need to create
    a dummy dataset (using, *e.g.*, :class:`aspecd.model.Zeros`) of given
    length and axes range. Based on that you can create your exponential:

    .. code-block:: yaml

       - kind: model
         type: Zeros
         properties:
           parameters:
             shape: 1001
             range: [-5, 5]
         result: dummy

       - kind: model
         type: Exponential
         from_dataset: dummy
         result: exponential

    Of course, if you start with an existing dataset (*e.g.*, loaded from
    some real data), you could use the label to this dataset directly in
    ``from_dataset``, without needing to create a dummy dataset first.

    Of course, you can control all parameters (prefactor, rate) explicitly:

    .. code-block:: yaml

       - kind: model
         type: Exponential
         properties:
           parameters:
             prefactor: 42
             rate: 4.2
         from_dataset: dummy
         result: exponential

    This would create an exponential with a prefactor of 42 (*i.e.* the
    intercept) and a rate of 4.2.

    .. versionadded:: 0.3

    """

    def __init__(self):
        super().__init__()
        self.description = "Exponential function"
        self.parameters["prefactor"] = 1
        self.parameters["rate"] = 1

    def _perform_task(self):
        x = np.asarray(self.variables[0])  # pylint: disable=invalid-name
        prefactor = self.parameters["prefactor"]
        rate = self.parameters["rate"]
        exponential = prefactor * np.exp(rate * x)
        self._dataset.data.data = exponential
